from abc import ABC, abstractmethod
from typing import Tuple, Dict, Any, Optional, List

class AbstractModelManager(ABC):
    """Abstract base class for model management."""

    @abstractmethod
    def __init__(
        self,
        model_id: str,
        internal_server_type: str,
        internal_port: int,
        internal_host: str,
        action_tracker: Any,
        num_model_instances: int = 1,
    ):
        """Initialize the model manager.

        Args:
            model_id: ID of the model.
            internal_server_type: Type of internal server.
            internal_port: Internal port number.
            internal_host: Internal host address.
            action_tracker: Tracker for monitoring actions.
            num_model_instances: Number of model instances to create.
        """
        pass

    @abstractmethod
    def get_model(self) -> Any:
        """Get a model instance for inference."""
        pass

    @abstractmethod
    def inference(
        self,
        input1: Any,
        input2: Optional[Any] = None,
        extra_params: Optional[Dict[str, Any]] = None,
        stream_key: Optional[str] = None,
        stream_info: Optional[Dict[str, Any]] = None,
        input_hash: Optional[str] = None,
    ) -> Tuple[Any, bool]:
        """Perform single inference."""
        pass

    @abstractmethod
    def batch_inference(
        self,
        input1: List[Any],
        input2: Optional[List[Any]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
        stream_key: Optional[str] = None,
        stream_info: Optional[Dict[str, Any]] = None,
        input_hash: Optional[str] = None,
    ) -> Tuple[List[Any], bool]:
        """Perform batch inference."""
        pass


import numpy as np
from typing import List, Dict, Tuple, Optional
from dataclasses import dataclass, field
import cv2
import requests
from scipy.special import softmax
try:
    from transformers import CLIPProcessor
    from PIL import Image
    import onnxruntime as ort
except:
    print("Unable to import onnxruntime/transformers/scipy/PIL")

def load_model_from_checkpoint(checkpoint_path,local_path):
    """
    Load a model from checkpoint URL
    """
    try:
        print(f"Loading model from checkpoint: {checkpoint_path}")
        
        # Check if checkpoint is a URL
        if checkpoint_path.startswith(('http://', 'https://')):
            # Download checkpoint from URL
            response = requests.get(checkpoint_path, timeout = (30,200))
            if response.status_code == 200:
                with open(local_path, 'wb') as f:
                    f.write(response.content)
                checkpoint_path = local_path
                print(f"Downloaded checkpoint to {local_path}")
            else:
                print(f"Failed to download checkpoint from {checkpoint_path}")
                return None
        
        # Load the model from the checkpoint
        model = ort.InferenceSession(checkpoint_path, providers=["CUDAExecutionProvider"])
        print(f"{local_path} Model loaded successfully from checkpoint")
        return model
        
    except Exception as e:
        print(f"Error loading model from checkpoint: {e}")
        return None



class ClipProcessor:
    def __init__(self,
                 image_model_path: str = 'https://s3.us-west-2.amazonaws.com/testing.resources/datasets/clip_image.onnx',
                 text_model_path: str = 'https://s3.us-west-2.amazonaws.com/testing.resources/datasets/clip_text.onnx',
                 processor_dir: str = './clip_processor',
                 providers: Optional[List[str]] = None):
        #test
        self.color_category: List[str] = ["black", "white", "yellow", "gray", "red", "blue", "light blue",
        "green", "brown"]

        self.image_url: str = image_model_path
        self.text_url: str = text_model_path
        self.processor_path: str = processor_dir

        self.image_sess = load_model_from_checkpoint(self.image_url,"clip_image.onnx")
        self.text_sess = load_model_from_checkpoint(self.text_url,"clip_text.onnx")

        self.processor = CLIPProcessor.from_pretrained(self.processor_path)

        tok = self.processor.tokenizer(self.color_category, padding=True, return_tensors="np")
        ort_inputs_text = {
            "input_ids": tok["input_ids"].astype(np.int64),
            "attention_mask": tok["attention_mask"].astype(np.int64)
        }
        text_out = self.text_sess.run(["text_embeds"], ort_inputs_text)[0].astype(np.float32)
        self.text_embeds = text_out / np.linalg.norm(text_out, axis=-1, keepdims=True)

        sample = self.processor(images=np.zeros((224, 224, 3), dtype=np.uint8), return_tensors="np")
        self.pixel_template = sample["pixel_values"].astype(np.float32)
        self.min_box_size = 32
        self.max_batch = 32
        self.frame_skip = 2
        self.batch_pixels = np.zeros((self.max_batch, *self.pixel_template.shape[1:]), dtype=np.float32)

        self.records: Dict[int, Dict[str, float]] = {}
        self.frame_idx = 0
        self.processed_frames = 0


    def process_color_in_frame(self, detections, input_bytes, zones: Optional[Dict[str, List[List[float]]]], stream_info):
        boxes = []
        tracked_ids: List[int] = []
        frame_number: Optional[int] = None
        # print(detections)
        self.frame_idx+=1
        nparr = np.frombuffer(input_bytes, np.uint8)        # convert bytes to numpy array
        image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)      # decode image

            # Step 2: Convert PIL → NumPy array
        frame = np.array(image)
        if stream_info:
            input_settings = stream_info.get("input_settings", {})
            start_frame = input_settings.get("start_frame")
            end_frame = input_settings.get("end_frame")
            if start_frame is not None and end_frame is not None and start_frame == end_frame:
                frame_number = start_frame

        for det in detections:
            bbox = det.get('bounding_box')
            tid = det.get('track_id')
            zones = zones if zones else {}
            for z_name, zone_polygon in zones.items():
                if self._is_in_zone(bbox, zone_polygon):
                    w = bbox['xmax'] - bbox['xmin']
                    h =  bbox['ymax'] - bbox['ymin']
                    if w >= self.min_box_size and h >= self.max_batch:
                        boxes.append(bbox)
                        tracked_ids.append(tid)
        # print(boxes)
        # print(tracked_ids)
        if not boxes:
          print(f"Frame {self.frame_idx}: No cars in zone")
          self.processed_frames += 1
          # print(f"Frame {frame_idx} processedms\n")
          return

        # print(boxes)
        # print(tracked_ids)
        crops_for_model = []
        map_trackidx_to_cropidx = []
        for i,(bbox, tid) in enumerate(zip(boxes, tracked_ids)):
            last_rec = self.records.get(tid)
            should_classify = False
            if last_rec is None:
                should_classify = True
            else:
                if (self.frame_idx - last_rec.get("last_classified_frame", -999)) >= self.frame_skip:
                    should_classify = True
            if should_classify:
                x1, y1, x2, y2 = bbox['xmin'], bbox['ymin'], bbox['xmax'], bbox['ymax']
                # crop safely
                y1c, y2c = max(0, y1), min(frame.shape[0], y2)
                x1c, x2c = max(0, x1), min(frame.shape[1], x2)
                if y2c - y1c <= 0 or x2c - x1c <= 0:
                    continue
                crop = cv2.cvtColor(frame[y1c:y2c, x1c:x2c], cv2.COLOR_BGR2RGB)
                pil_img = Image.fromarray(crop).resize((224, 224))
                map_trackidx_to_cropidx.append((tid, len(crops_for_model)))
                crops_for_model.append(pil_img)
        # print(crops_for_model)
        # print(map_trackidx_to_cropidx)

        if crops_for_model:
            record = {}
            img_embeds = self.run_image_onnx_on_crops(crops_for_model)  # [N, D]
            # compute similarity with text_embeds (shape [num_labels, D])
            sims = img_embeds @ self.text_embeds.T  # [N, num_labels]
            # convert to probs
            probs = np.exp(sims) / np.exp(sims).sum(axis=-1, keepdims=True)  # softmax numerically simple
            # print(probs)

            # assign back to corresponding tracks
            for (tid, crop_idx) in map_trackidx_to_cropidx:
                prob = probs[crop_idx]
                # print(prob)
                best_idx = int(np.argmax(prob))
                best_label = self.color_category[best_idx]
                # print(best_label)
                best_score = float(prob[best_idx])
                # print(best_score)

                rec = self.records.get(tid)
                # if rec is None:
                record[tid] = {
                    "frame": self.frame_idx,
                    "color": best_label,
                    "confidence": best_score,
                    "track_id": tid,
                    "last_classified_frame": self.frame_idx,
                }
                # else:
                #     # update only if confidence improves
                #     if best_score > rec["confidence"]:
                #         rec["color"] = best_label
                #         rec["confidence"] = best_score
                #         rec["frame"] = self.frame_idx
                #     rec["last_classified_frame"] = self.frame_idx


        return record


    def run_image_onnx_on_crops(self, crops):
        valid_crops = []
        for i, crop in enumerate(crops):
            if isinstance(crop, Image.Image):  # PIL.Image
                crop = np.array(crop)
            if not isinstance(crop, np.ndarray):
                print(f"Skipping crop {i}: not a numpy array ({type(crop)})")
                continue
            if crop.size == 0:
                print(f"Skipping crop {i}: empty array")
                continue

            try:
                crop_resized = cv2.resize(crop, (224, 224), interpolation=cv2.INTER_LINEAR)
                valid_crops.append(crop_resized)
            except Exception as e:
                print(f"Skipping crop {i}: resize failed ({e})")

        if not valid_crops:
            print("⚠️ No valid crops to process")
            return np.zeros((0, self.text_embeds.shape[-1]), dtype=np.float32)

        # Convert all valid crops at once
        pixel_values = self.processor(images=valid_crops, return_tensors="np")["pixel_values"]
        n = pixel_values.shape[0]
        self.batch_pixels[:n] = pixel_values

        ort_inputs = {"pixel_values": self.batch_pixels[:n]}
        img_out = self.image_sess.run(["image_embeds"], ort_inputs)[0].astype(np.float32)

        return img_out / np.linalg.norm(img_out, axis=-1, keepdims=True)


    def _is_in_zone(self, bbox, polygon: List[List[float]]) -> bool:
        if not polygon:
            return False
        # print(bbox)
        x1, y1, x2, y2 = bbox['xmin'], bbox['ymin'], bbox['xmax'], bbox['ymax']
        # print(x1,x2,y1,y2)
        # print(type(x1))
        # print(polygon)
        cx, cy = int((x1 + x2) / 2), int((y1 + y2) / 2)
        polygon = np.array(polygon, dtype=np.int32)
        return cv2.pointPolygonTest(polygon, (cx, cy), False) >= 0





import logging
import numpy as np
import requests
import time
from typing import Tuple, Dict, Any, Optional, List, Union
from simplified_triton_utils import TritonServer, TritonInference
import logging
import time
import requests
from typing import Union, List

class TritonModelManager:
    """Model manager for Triton Inference Server, aligned with pipeline and inference interface."""

    def __init__(
        self,
        model_name: str,
        model_path: str,
        runtime_framework: str,
        internal_server_type: str,
        internal_port: int,
        internal_host: str,
        input_size: Union[int, List[int]] = 640, # Priority Obj det
        num_classes: int = 10,
        num_model_instances: int = 1,
        use_dynamic_batching: bool = False,
        max_batch_size: int = 8,

        is_yolo: bool = False,
        use_trt_accelerator: bool = False,
    ):
        try:
            if internal_server_type not in ["rest", "grpc"]:
                logging.warning(f"Invalid internal_server_type '{internal_server_type}', defaulting to 'rest'")

            self.internal_server_type = internal_server_type
            self.internal_port = internal_port
            self.internal_host = internal_host
            self.use_dynamic_batching = use_dynamic_batching
            self.max_batch_size = max_batch_size
            self.model_instances = []
            self._round_robin_counter = 0

            self.triton_server = TritonServer(
                model_name=model_name,
                model_path=model_path,
                runtime_framework=runtime_framework,
                input_size=input_size,
                num_classes=num_classes,
                dynamic_batching=use_dynamic_batching,
                num_model_instances=num_model_instances,
                max_batch_size=max_batch_size,
                connection_protocol=internal_server_type,
                is_yolo=is_yolo,
                use_trt_accelerator=use_trt_accelerator,
            )
            
            logging.info(f"Starting Triton server on {internal_host}:{internal_port}...")
            self.triton_server_process = self.triton_server.setup(internal_port)

            logging.info("Waiting for Triton server to be ready...")
            self._wait_for_ready()

            for _ in range(num_model_instances):
                client = TritonInference(
                    server_type=self.triton_server.connection_protocol,
                    model_name=model_name,
                    internal_port=internal_port,
                    internal_host=internal_host,
                    runtime_framework=self.triton_server.runtime_framework,
                    is_yolo = self.triton_server.is_yolo,
                    input_size=input_size,
                    # action_tracker=self.triton_server.action_tracker,
                )
                self.model_instances.append(client)
            
            logging.info(f"Initialized TritonModelManager with {num_model_instances} client instances, protocol: {self.triton_server.connection_protocol}")
            
        except Exception as e:
            logging.error(f"Failed to initialize TritonModelManager: {str(e)}", exc_info=True)
            raise

    def _wait_for_ready(self):
        """Wait for Triton server to be ready with fixed retries and 5s sleep."""
        max_attempts = 30  # 150 seconds wait time
        for attempt in range(max_attempts):
            try:
                if self.internal_server_type == "rest":
                    response = requests.get(
                        f"http://{self.internal_host}:{self.internal_port}/v2/health/ready",
                        timeout=5
                    )
                    if response.status_code == 200:
                        logging.info("=========  Triton server is ready (REST) =========")
                        break
                    else:
                        logging.info(f"Attempt {attempt + 1}/{max_attempts} - server not ready, retrying in 5 seconds...")
                        time.sleep(5)

                else:  # gRPC
                    try:
                        import tritonclient.grpc as grpcclient
                    except ImportError:
                        grpcclient = None

                    if grpcclient is None:
                        raise ImportError("tritonclient.grpc required for gRPC")

                    with grpcclient.InferenceServerClient(f"{self.internal_host}:{self.internal_port}") as client:
                        if client.is_server_ready():
                            logging.info("=========  Triton server is ready (gRPC) =========")
                            break
                        else:
                            logging.info(f"Attempt {attempt + 1}/{max_attempts} - server not ready, retrying in 5 seconds...")
                            time.sleep(5)

            except Exception as e:
                if attempt < max_attempts - 1:
                    logging.info(f"Attempt {attempt + 1}/{max_attempts} failed, retrying in 5 seconds... (Error: {str(e)})")
                    time.sleep(5)
                else:
                    logging.error("Triton server failed to become ready after maximum attempts")
                    raise

            
    def get_model(self) -> Any:
        """Get a TritonInference client instance in round-robin fashion."""
        if not self.model_instances:
            logging.warning(f"No Triton clients available for model")
            return None
        model = self.model_instances[self._round_robin_counter % len(self.model_instances)]
        self._round_robin_counter = (self._round_robin_counter + 1) % len(self.model_instances)
        return model

    def inference(
        self,
        input1: Any,
        input2: Optional[Any] = None,
        extra_params: Optional[Dict[str, Any]] = None,
        stream_key: Optional[str] = None,
        stream_info: Optional[Dict[str, Any]] = None,
        input_hash: Optional[str] = None,
    ) -> Tuple[Any, bool]:
        """Perform synchronous single inference using TritonInference client.

        Args:
            input1: Primary input data (e.g., image bytes).
            input2: Secondary input data (optional, ignored for Triton).
            extra_params: Additional parameters for inference.
            stream_key: Stream key for caching.
            stream_info: Stream info for post-processing.
            input_hash: Input hash for caching.

        Returns:
            Tuple of (results, success_flag).
        """
        if input1 is None:
            raise ValueError("Input data cannot be None")
        try:
            client = self.get_model()
            if not client:
                raise RuntimeError("No Triton client available")
            results = client.inference(input1)
            results = client.format_response(results)
            return results, True
        except Exception as e:
            logging.error(f"Triton sync inference failed for: {str(e)}", exc_info=True)
            return None, False

    async def async_inference(
        self,
        input1: Union[bytes, np.ndarray],
        input2: Optional[Any] = None,
        extra_params: Optional[Dict[str, Any]] = None,
        stream_key: Optional[str] = None,
        stream_info: Optional[Dict[str, Any]] = None,
        input_hash: Optional[str] = None,
    ) -> Tuple[Any, bool]:
        """Perform asynchronous single inference using TritonInference client.
        Args:
            input1: Primary input data (Image bytes or numpy array).
            input2: Secondary input data (optional, ignored for Triton).
            extra_params: Additional parameters for inference.
            stream_key: Stream key for caching.
            stream_info: Stream info for post-processing.
            input_hash: Input hash for caching.
        Returns:
            Tuple of (results, success_flag).
        """
        

        if input1 is None:
            logging.error("Input data cannot be None")
            raise ValueError("Input data cannot be None")
        try:
            client = self.get_model()
            if not client:
                logging.error("No Triton client available")
                raise RuntimeError("No Triton client available")
            results = await client.async_inference(input1)
            results = client.format_response(results)
            logging.info(f"Async inference result: {results}")
            return results, True
        except Exception as e:
            logging.error(f"Triton async inference failed: {e}")
            return {"error": str(e), "predictions": None}, False

    def batch_inference(
        self,
        input1: List[Any],
        input2: Optional[List[Any]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
        stream_key: Optional[str] = None,
        stream_info: Optional[Dict[str, Any]] = None,
        input_hash: Optional[str] = None,
    ) -> Tuple[List[Any], bool]:
        """Perform synchronous batch inference using TritonInference client.

        Args:
            input1: List of primary input data (e.g., image bytes).
            input2: List of secondary input data (optional, ignored for Triton).
            extra_params: Additional parameters for inference.
            stream_key: Stream key for caching.
            stream_info: Stream info for post-processing.
            input_hash: Input hash for caching.

        Returns:
            Tuple of (results_list, success_flag).
        """
        if not input1:
            raise ValueError("Batch input cannot be None")
        try:
            client = self.get_model()
            if not client:
                raise RuntimeError("No Triton client available")
            results = []

            if self.use_dynamic_batching:
                input_array = self._preprocess_batch_inputs(input1, client)
                batch_results = client.inference(input_array)
                results = self._split_batch_results(batch_results, len(input1))
            else:
                for inp in input1:
                    result = client.inference(inp)
                    results.append(result)

            results = [client.format_response(result) for result in results]
            return results, True
        except Exception as e:
            logging.error(f"Triton sync batch inference failed for: {str(e)}", exc_info=True)
            return None, False

    async def async_batch_inference(
        self,
        input1: List[Any],
        input2: Optional[List[Any]] = None,
        extra_params: Optional[Dict[str, Any]] = None,
        stream_key: Optional[str] = None,
        stream_info: Optional[Dict[str, Any]] = None,
        input_hash: Optional[str] = None,
    ) -> Tuple[List[Any], bool]:
        """Perform asynchronous batch inference using TritonInference client.

        Args:
            input1: List of primary input data (e.g., image bytes).
            input2: List of secondary input data (optional, ignored for Triton).
            extra_params: Additional parameters for inference.
            stream_key: Stream key for caching.
            stream_info: Stream info for post-processing.
            input_hash: Input hash for caching.

        Returns:
            Tuple of (results_list, success_flag).
        """
        if not input1:
            raise ValueError("Batch input cannot be None")
        try:
            client = self.get_model()
            if not client:
                raise RuntimeError("No Triton client available")
            results = []

            if self.use_dynamic_batching:
                input_array = self._preprocess_batch_inputs(input1, client)
                batch_results = await client.async_inference(input_array)
                split_results = self._split_batch_results(batch_results, len(input1))
                results = [client.format_response(r) for r in split_results]
            else:
                for inp in input1:
                    res = await client.async_inference(inp)
                    results.append(client.format_response(res))

            return results, True
        except Exception as e:
            logging.error(f"Triton async batch inference failed for: {str(e)}", exc_info=True)
            return None, False

    def _preprocess_batch_inputs(self, input1: List[Any], client: TritonInference) -> np.ndarray:
        """Preprocess batch inputs for Triton dynamic batching.

        Args:
            input1: List of input data (e.g., image bytes).
            client: TritonInference client for shape and data type information.

        Returns:
            Preprocessed NumPy array for batch inference.
        """
        try:
            batch_inputs = []
            for inp in input1:
                arr = client._preprocess_input(inp)

                if arr.ndim == 4 and arr.shape[0] == 1:
                    arr = np.squeeze(arr, axis=0)

                if arr.ndim != 3:
                    logging.warning(f"Unexpected input shape {arr.shape}, expected (C,H,W) after preprocessing")

                batch_inputs.append(arr)

            # Stack into final batch (B, C, H, W)
            stacked = np.stack(batch_inputs, axis=0)
            # Ensure C-contiguous (important for Triton)
            return np.ascontiguousarray(stacked)

        except Exception as e:
            logging.error(f"Failed to preprocess batch inputs: {str(e)}", exc_info=True)
            raise


    def _split_batch_results(self, batch_results: np.ndarray, batch_size: int) -> List[Any]:
        """Split batch results into individual results.

        Args:
            batch_results: NumPy array of batch inference results.
            batch_size: Number of inputs in the batch.

        Returns:
            List of individual results.
        """
        try:
            if batch_results.ndim == 1:
                return [batch_results] * batch_size
            return [batch_results[i] for i in range(batch_size)]
        except Exception as e:
            logging.error(f"Failed to split batch results: {str(e)}", exc_info=True)
            raise


"""Module providing metrics_calculator functionality."""

import torch
import torchvision
from sklearn.preprocessing import label_binarize
from sklearn.metrics import (
    precision_score,
    recall_score,
    f1_score,
    matthews_corrcoef,
    roc_auc_score,
    average_precision_score,
    cohen_kappa_score,
    log_loss,
)


def get_object_detection_evaluation_results(
    split,
    all_outputs,
    all_targets,
    index_to_labels,
):
    """
    Calculate and format object detection evaluation metrics.

    Args:
        split: Dataset split type (e.g., 'train', 'val', 'test')
        all_outputs: Model predictions
        all_targets: Ground truth annotations
        index_to_labels: Mapping from class indices to label names

    Returns:
        List of dictionaries containing formatted metrics
    """
    results = []
    (
        precision_class,
        recall_class,
        f1_score_class,
    ) = calculate_detection_metrics(
        all_outputs,
        all_targets,
        len(index_to_labels),
    )
    (
        mAP_50_class,
        mAP_75_class,
        mAP_90_class,
        mAP_50_95_class,
    ) = calculate_mAP_metrics(
        all_outputs,
        all_targets,
        len(index_to_labels),
    )
    (
        mAR_50_class,
        mAR_75_class,
        mAR_90_class,
        mAR_50_95_class,
    ) = calculate_mAR_metrics(
        all_outputs,
        all_targets,
        len(index_to_labels),
    )
    results.extend(
        [
            {
                "category": "all",
                "splitType": split,
                "metricName": "precision",
                "metricValue": float(precision_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "recall",
                "metricValue": float(recall_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "f1_score",
                "metricValue": float(f1_score_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "mAP@50",
                "metricValue": float(mAP_50_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "mAP@75",
                "metricValue": float(mAP_75_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "mAP@50-95",
                "metricValue": float(mAP_50_95_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "mAP@90",
                "metricValue": float(mAP_90_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "mAR@50",
                "metricValue": float(mAR_50_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "mAR@75",
                "metricValue": float(mAR_75_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "mAR@50-95",
                "metricValue": float(mAR_50_95_class.mean().item()),
            },
            {
                "category": "all",
                "splitType": split,
                "metricName": "mAR@90",
                "metricValue": float(mAR_90_class.mean().item()),
            },
        ]
    )
    metrics_per_class = {
        "precision": precision_class,
        "recall": recall_class,
        "f1_score": f1_score_class,
        "AP@50": mAP_50_class,
        "AP@75": mAP_75_class,
        "AP@90": mAP_90_class,
        "AP@50-95": mAP_50_95_class,
        "AR@50": mAR_50_class,
        "AR@75": mAR_75_class,
        "AR@90": mAR_90_class,
        "AR@50-95": mAR_50_95_class,
    }
    for i in range(len(index_to_labels)):
        for (
            metric_name,
            metric_value,
        ) in metrics_per_class.items():
            if metric_value[i].isnan().item():
                metric_value[i] = torch.tensor(0.0)
            results.append(
                {
                    "category": index_to_labels[str(i)],
                    "splitType": split,
                    "metricName": metric_name,
                    "metricValue": float(metric_value[i].item()),
                }
            )
    return results


def calculate_mAP_metrics(outputs, targets, num_classes):
    mAP = torch.zeros(num_classes)
    mAP_50 = torch.zeros(num_classes)
    mAP_75 = torch.zeros(num_classes)
    mAP_50_95 = torch.zeros(num_classes)
    mAP_90 = torch.zeros(num_classes)
    iou_thresholds_50_95 = torch.linspace(0.5, 0.95, 10)
    iou_threshold_standard = 0.5
    for label in range(num_classes):
        all_predictions, all_targets = collect_predictions_and_targets(outputs, targets, label)
        mAP[label] = calculate_ap(
            all_predictions,
            all_targets,
            iou_threshold_standard,
        )
        mAP_50[label] = calculate_ap(all_predictions, all_targets, 0.5)
        mAP_75[label] = calculate_ap(all_predictions, all_targets, 0.75)
        mAP_90[label] = calculate_ap(all_predictions, all_targets, 0.9)
        ap_sum_50_95 = sum(
            calculate_ap(
                all_predictions,
                all_targets,
                iou_threshold,
            )
            for iou_threshold in iou_thresholds_50_95
        )
        mAP_50_95[label] = ap_sum_50_95 / len(iou_thresholds_50_95)
    return mAP_50, mAP_75, mAP_90, mAP_50_95


def collect_predictions_and_targets(outputs, targets, label):
    all_predictions = []
    all_targets = []
    for output, target in zip(outputs, targets):
        pred_boxes = output["boxes"][output["labels"] == label]
        pred_scores = output["scores"][output["labels"] == label]
        target_boxes = target["boxes"][target["labels"] == label]
        all_predictions.append((pred_boxes, pred_scores))
        all_targets.append(target_boxes)
    return all_predictions, all_targets


def calculate_ap(predictions, targets, iou_threshold):
    """
    Calculate Average Precision for a single class at a specific IoU threshold
    """
    total_gt = sum(len(t) for t in targets)
    if total_gt == 0:
        return torch.tensor(0.0)
    all_predictions = [
        (box, score)
        for pred_boxes, pred_scores in predictions
        for box, score in zip(pred_boxes, pred_scores)
    ]
    if not all_predictions:
        return torch.tensor(0.0)
    all_predictions.sort(key=lambda x: x[1], reverse=True)
    matched_targets = {i: [] for i in range(len(targets))}
    precisions, recalls = calculate_precision_recall(
        all_predictions,
        targets,
        matched_targets,
        iou_threshold,
        0,
        0,
        total_gt,
    )
    if not precisions:
        return torch.tensor(0.0)
    precisions = torch.tensor(precisions)
    recalls = torch.tensor(recalls)
    for i in range(len(precisions) - 1, 0, -1):
        precisions[i - 1] = max(precisions[i - 1], precisions[i])
    ap = 0.0
    for i in range(len(precisions)):
        if i == 0:
            ap += precisions[i] * recalls[i]
        else:
            ap += precisions[i] * (recalls[i] - recalls[i - 1])
    return torch.tensor(ap)


def calculate_precision_recall(
    all_predictions,
    targets,
    matched_targets,
    iou_threshold,
    total_tp,
    total_fp,
    total_gt,
):
    precisions = []
    recalls = []
    for pred_box, _ in all_predictions:
        max_iou, best_match, best_target_idx = find_best_match(
            pred_box,
            targets,
            matched_targets,
            iou_threshold,
        )
        if best_match is not None:
            matched_targets[best_target_idx].append(best_match)
            total_tp += 1
        else:
            total_fp += 1
        precision = total_tp / (total_tp + total_fp)
        recall = total_tp / total_gt if total_gt > 0 else 0
        precisions.append(precision)
        recalls.append(recall)
    return precisions, recalls


def calculate_mAR_metrics(outputs, targets, num_classes):
    """
    Calculate average recall metrics for object detection.
    """
    torch.zeros(num_classes)
    mAR_50 = torch.zeros(num_classes)
    mAR_75 = torch.zeros(num_classes)
    mAR_50_95 = torch.zeros(num_classes)
    mAR_90 = torch.zeros(num_classes)
    iou_thresholds_50_95 = torch.linspace(0.5, 0.95, 10)
    for label in range(num_classes):
        all_predictions, all_targets = collect_predictions_and_targets(outputs, targets, label)
        mAR_50[label] = calculate_recall_at_iou(all_predictions, all_targets, 0.5)
        mAR_75[label] = calculate_recall_at_iou(all_predictions, all_targets, 0.75)
        mAR_90[label] = calculate_recall_at_iou(all_predictions, all_targets, 0.9)
        mAR_50_95[label] = sum(
            calculate_recall_at_iou(
                all_predictions,
                all_targets,
                iou_threshold,
            )
            for iou_threshold in iou_thresholds_50_95
        ) / len(iou_thresholds_50_95)
    return mAR_50, mAR_75, mAR_90, mAR_50_95


def calculate_recall_at_iou(predictions, targets, iou_threshold):
    """Calculate recall at a specific IoU threshold for a single class."""
    total_gt = sum(len(t) for t in targets)
    if total_gt == 0:
        return torch.tensor(0.0)
    total_tp = 0
    matched_targets = {i: [] for i in range(len(targets))}
    for pred_boxes, pred_scores in predictions:
        for pred_box, _ in zip(pred_boxes, pred_scores):
            (
                max_iou,
                best_match,
                best_target_idx,
            ) = find_best_match(
                pred_box,
                targets,
                matched_targets,
                iou_threshold,
            )
            if best_match is not None:
                matched_targets[best_target_idx].append(best_match)
                total_tp += 1
    recall = total_tp / total_gt
    return torch.tensor(recall)


def find_best_match(
    pred_box,
    targets,
    matched_targets,
    iou_threshold,
):
    max_iou = iou_threshold
    best_match = None
    best_target_idx = None
    for target_idx, target_boxes in enumerate(targets):
        if len(target_boxes) == 0:
            continue
        unmatched_indices = [
            i for i in range(len(target_boxes)) if i not in matched_targets[target_idx]
        ]
        if not unmatched_indices:
            continue
        target_boxes_unmatched = target_boxes[unmatched_indices]
        ious = torchvision.ops.box_iou(
            pred_box.unsqueeze(0),
            target_boxes_unmatched,
        )
        max_iou_for_target, max_idx = ious.max(dim=1)
        if max_iou_for_target > max_iou:
            max_iou = max_iou_for_target
            best_match = unmatched_indices[max_idx]
            best_target_idx = target_idx
    return max_iou, best_match, best_target_idx


def calculate_detection_metrics(outputs, targets, num_classes):
    all_true_positives = torch.zeros(num_classes)
    all_false_positives = torch.zeros(num_classes)
    all_false_negatives = torch.zeros(num_classes)
    iou_threshold = 0.5
    for output, target in zip(outputs, targets):
        for label in range(num_classes):
            pred_boxes = output["boxes"][output["labels"] == label]
            output["scores"][output["labels"] == label]
            target_boxes = target["boxes"][target["labels"] == label]
            if len(target_boxes) == 0:
                if len(pred_boxes) > 0:
                    all_false_positives[label] += len(pred_boxes)
                continue
            if len(pred_boxes) == 0:
                all_false_negatives[label] += len(target_boxes)
                continue
            ious = torchvision.ops.box_iou(pred_boxes, target_boxes)
            matched_pred_indices = set()
            for target_idx in range(len(target_boxes)):
                max_iou, pred_idx = ious[:, target_idx].max(dim=0)
                if max_iou >= iou_threshold and pred_idx.item() not in matched_pred_indices:
                    all_true_positives[label] += 1
                    matched_pred_indices.add(pred_idx.item())
                else:
                    all_false_negatives[label] += 1
            all_false_positives[label] += len(pred_boxes) - len(matched_pred_indices)
    precision_class = all_true_positives / (all_true_positives + all_false_positives + 1e-10)
    recall_class = all_true_positives / (all_true_positives + all_false_negatives + 1e-10)
    f1_score_class = (
        2 * (precision_class * recall_class) / (precision_class + recall_class + 1e-10)
    )
    return (
        precision_class,
        recall_class,
        f1_score_class,
    )


def get_classification_evaluation_results(split_type, outputs, targets, index_to_labels):
    predictions = torch.argmax(outputs, dim=1)
    results = []
    acc1 = accuracy(outputs, targets, topk=(1,))[0]
    acc5 = (
        accuracy(outputs, targets, topk=(5,))[0]
        if len(index_to_labels) >= 5
        else torch.tensor([1])
    )
    predictions_cpu = predictions.cpu()
    targets_cpu = targets.cpu()
    macro_metrics = {
        "precision": precision_score(
            targets_cpu,
            predictions_cpu,
            average="macro",
            zero_division=0,
        ),
        "recall": recall_score(
            targets_cpu,
            predictions_cpu,
            average="macro",
            zero_division=0,
        ),
        "f1_score": f1_score(
            targets_cpu,
            predictions_cpu,
            average="macro",
            zero_division=0,
        ),
    }
    micro_metrics = {
        "precision": precision_score(
            targets_cpu,
            predictions_cpu,
            average="micro",
            zero_division=0,
        ),
        "recall": recall_score(
            targets_cpu,
            predictions_cpu,
            average="micro",
            zero_division=0,
        ),
        "f1_score": f1_score(
            targets_cpu,
            predictions_cpu,
            average="micro",
            zero_division=0,
        ),
    }
    weighted_metrics = {
        "precision": precision_score(
            targets_cpu,
            predictions_cpu,
            average="weighted",
            zero_division=0,
        ),
        "recall": recall_score(
            targets_cpu,
            predictions_cpu,
            average="weighted",
            zero_division=0,
        ),
        "f1_score": f1_score(
            targets_cpu,
            predictions_cpu,
            average="weighted",
            zero_division=0,
        ),
    }
    additional_metrics = {
        "acc@1": acc1.item(),
        "acc@5": acc5.item(),
        "MCC": calculate_mcc(predictions, targets),
        "AUC-ROC": calculate_auc_roc(outputs, targets, len(index_to_labels)),
        "AUC-PR": calculate_auc_pr(outputs, targets, len(index_to_labels)),
        "Cohen's Kappa": calculate_cohen_kappa(predictions, targets),
        "log_loss": calculate_log_loss(outputs, targets),
        "specificity": specificity_all(outputs, targets),
        "micro_precision": micro_metrics["precision"],
        "micro_recall": micro_metrics["recall"],
        "micro_f1_score": micro_metrics["f1_score"],
        "macro_precision": macro_metrics["precision"],
        "macro_recall": macro_metrics["recall"],
        "macro_f1_score": macro_metrics["f1_score"],
        "weighted_precision": weighted_metrics["precision"],
        "weighted_recall": weighted_metrics["recall"],
        "weighted_f1_score": weighted_metrics["f1_score"],
    }
    for (
        metric_name,
        value,
    ) in additional_metrics.items():
        results.append(
            {
                "category": "all",
                "splitType": split_type,
                "metricName": metric_name,
                "metricValue": float(value),
            }
        )
    metrics_per_class = {
        "precision": precision,
        "f1_score": f1_score_per_class,
        "recall": recall,
        "specificity": specificity,
        "acc@1": accuracy_per_class,
    }
    for (
        metric_name,
        metric_func,
    ) in metrics_per_class.items():
        class_metrics = metric_func(outputs, targets)
        for (
            class_idx,
            value,
        ) in class_metrics.items():
            results.append(
                {
                    "category": index_to_labels[str(class_idx)],
                    "splitType": split_type,
                    "metricName": metric_name,
                    "metricValue": float(value),
                }
            )
    return results


def calculate_metrics(output, target):
    """
    Calculate true positives, true negatives, false positives, and false negatives for a
        multi-class classification.
    """
    _, pred = output.max(1)
    pred = pred.cpu()
    target = target.cpu()
    num_classes = output.size(1)
    true_positives = torch.zeros(num_classes)
    true_negatives = torch.zeros(num_classes)
    false_positives = torch.zeros(num_classes)
    false_negatives = torch.zeros(num_classes)
    for class_idx in range(num_classes):
        true_positives[class_idx] = ((pred == class_idx) & (target == class_idx)).sum().item()
        true_negatives[class_idx] = ((pred != class_idx) & (target != class_idx)).sum().item()
        false_positives[class_idx] = ((pred == class_idx) & (target != class_idx)).sum().item()
        false_negatives[class_idx] = ((pred != class_idx) & (target == class_idx)).sum().item()
    return (
        true_positives,
        true_negatives,
        false_positives,
        false_negatives,
    )


def accuracy_per_class(output, target):
    tp, tn, fp, fn = calculate_metrics(output, target)
    accuracy_per_class = {}
    for class_idx in range(output.size(1)):
        total = tp[class_idx] + tn[class_idx] + fp[class_idx] + fn[class_idx]
        accuracy = (tp[class_idx] + tn[class_idx]) / total if total > 0 else 0.0
        accuracy_per_class[class_idx] = accuracy
    return accuracy_per_class


def specificity_all(output, target):
    _, tn, fp, _ = calculate_metrics(output, target)
    total_tn = tn.sum().item()
    total_fp = fp.sum().item()
    return total_tn / (total_tn + total_fp + 1e-10)


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = min(max(topk), output.size(1))
        batch_size = target.size(0)
        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))
        res = []
        for k in topk:
            k = min(k, maxk)
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


def precision(output, target):
    tp, _, fp, _ = calculate_metrics(output, target)
    precision_per_class = {}
    for class_idx in range(output.size(1)):
        denominator = tp[class_idx] + fp[class_idx]
        precision_per_class[class_idx] = tp[class_idx] / denominator if denominator > 0 else 0.0
    return precision_per_class


def recall(output, target):
    tp, _, _, fn = calculate_metrics(output, target)
    recall_per_class = {}
    for class_idx in range(output.size(1)):
        denominator = tp[class_idx] + fn[class_idx]
        recall_per_class[class_idx] = tp[class_idx] / denominator if denominator > 0 else 0.0
    return recall_per_class


def f1_score_per_class(output, target):
    precision_scores = precision(output, target)
    recall_scores = recall(output, target)
    f1_scores = {}
    for class_idx in range(output.size(1)):
        p = precision_scores[class_idx]
        r = recall_scores[class_idx]
        denominator = p + r
        f1_scores[class_idx] = 2 * (p * r) / denominator if denominator > 0 else 0.0
    return f1_scores


def specificity(output, target):
    _, tn, fp, _ = calculate_metrics(output, target)
    specificity_per_class = {}
    for class_idx in range(output.size(1)):
        denominator = tn[class_idx] + fp[class_idx]
        specificity_per_class[class_idx] = tn[class_idx] / denominator if denominator > 0 else 0.0
    return specificity_per_class


def confusion_matrix_per_class(output, target):
    tp, tn, fp, fn = calculate_metrics(output, target)
    confusion_matrix = {}
    for class_idx in range(output.size(1)):
        confusion_matrix[class_idx] = [
            [
                int(tp[class_idx]),
                int(fp[class_idx]),
            ],
            [
                int(fn[class_idx]),
                int(tn[class_idx]),
            ],
        ]
    return confusion_matrix


def confusion_matrix(output, target):
    num_classes = output.size(1)
    confusion_matrix_overall = torch.zeros(
        (num_classes, num_classes),
        dtype=torch.int64,
    )
    _, predicted_classes = output.max(1)
    for i in range(target.size(0)):
        predicted_class = predicted_classes[i]
        true_class = target[i]
        confusion_matrix_overall[true_class][predicted_class] += 1
    return confusion_matrix_overall


def calculate_mcc(predictions, targets):
    try:
        return matthews_corrcoef(
            targets.cpu().numpy(),
            predictions.cpu().numpy(),
        )
    except Exception as e:
        print(e)
        return 0.0


def calculate_auc_roc(outputs, targets, num_classes):
    try:
        targets_one_hot = label_binarize(
            targets.cpu().numpy(),
            classes=range(num_classes),
        )
        probabilities = torch.nn.functional.softmax(outputs, dim=1).cpu().numpy()
        if num_classes == 2:
            return roc_auc_score(
                targets_one_hot,
                probabilities[:, 1],
            )
        else:
            return roc_auc_score(
                targets_one_hot,
                probabilities,
                average="macro",
                multi_class="ovr",
            )
    except Exception as e:
        print(e)
        return 0.0


def calculate_auc_pr(outputs, targets, num_classes):
    try:
        targets_one_hot = label_binarize(
            targets.cpu().numpy(),
            classes=range(num_classes),
        )
        probabilities = torch.nn.functional.softmax(outputs, dim=1).cpu().numpy()
        return average_precision_score(
            targets_one_hot,
            probabilities,
            average="macro",
        )
    except Exception as e:
        print(e)
        return 0.0


def calculate_cohen_kappa(predictions, targets):
    try:
        return cohen_kappa_score(
            targets.cpu().numpy(),
            predictions.cpu().numpy(),
        )
    except Exception as e:
        print(e)
        return 0.0


def calculate_log_loss(outputs, targets):
    try:
        probabilities = torch.nn.functional.softmax(outputs, dim=1).cpu().numpy()
        return log_loss(
            targets.cpu().numpy(),
            probabilities,
            labels=range(outputs.size(1)),
        )
    except Exception as e:
        print(e)
        return 0.0



"""Module providing triton_server functionality."""

import os
import zipfile
import subprocess
import tempfile
import asyncio
import shutil
import logging
import threading
import shlex
from typing import Tuple, Optional, Any, Dict, Union, List
import importlib.util
from matrice.docker_utils import pull_docker_image
from matrice.utils import dependencies_check

# TRITON_DOCKER_IMAGE = "nvcr.io/nvidia/tritonserver:24.08-py3"
TRITON_DOCKER_IMAGE = "nvcr.io/nvidia/tritonserver:23.08-py3"
BASE_PATH = "./model_repository"

# def dependencies_check(package_name: str) -> bool:
#     """Check if a package is installed."""
#     return importlib.util.find_spec(package_name) is not None


# def pull_docker_image(image_name: str) -> subprocess.Popen:
#     """Pull Docker image using subprocess."""
#     cmd = ["docker", "pull", image_name]
#     return subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s] %(message)s")

class TritonServer:
    def __init__(
        self,
        model_name: str,
        model_path: str,
        runtime_framework: str,
        input_size: Union[int, List[int]] = 224,
        num_classes: int = 10,
        dynamic_batching: bool = False,
        num_model_instances: int = 1,
        max_batch_size: int = 8,
        connection_protocol: str = "rest",
        is_yolo: bool = False,
        use_trt_accelerator: bool = False,
        **kwargs,
    ):
        """Initialize the Triton server.

        Args:
            model_name: Name of the model (used for Triton model repository).
            model_path: Path to the model file on the local filesystem.
            runtime_framework: Framework of the model ('onnx', 'pytorch', 'torchscript', 'yolo', 'tensorrt', 'openvino').
            input_size: Input size for the model (int for square images or [height, width]).
            num_classes: Number of output classes.
            dynamic_batching: Enable dynamic batching for the model.
            num_model_instances: Number of model instances to deploy.
            max_batch_size: Maximum batch size for inference.
            connection_protocol: Protocol for Triton server ('rest' or 'grpc').
            use_trt_accelerator: Enable TensorRT acceleration for inference.
        """
        if not dependencies_check("torch"):
            raise ImportError("PyTorch is required but not installed")
        import torch
        
        if not model_name:
            raise ValueError("model_name must be provided")
        if not model_path:
            raise ValueError("model_path must be provided")
        if not os.path.exists(model_path):
            raise FileNotFoundError(f"Model file not found at: {model_path}")

        logging.info("Initializing TritonServer")
        
        self.model_name = model_name
        self.model_path = os.path.abspath(model_path)
        self.runtime_framework = runtime_framework.lower()
        self.connection_protocol = connection_protocol.lower()
        
        if isinstance(input_size, (int, float)):
            self.input_size = [int(input_size), int(input_size)]
        elif isinstance(input_size, (list, tuple)):
            if len(input_size) == 2:  # (H, W)
                self.input_size = [int(input_size[0]), int(input_size[1])]
            elif len(input_size) == 3:  # (H, W, C) 
                self.input_size = [int(input_size[0]), int(input_size[1])]
            elif len(input_size) == 4:  # (N, C, H, W) 
                self.input_size = [int(input_size[-2]), int(input_size[-1])]
        else:
            logging.warning("Unexpected input_size length: %s, using default [224, 224]", input_size)
            self.input_size = [224, 224]

        self.num_classes = num_classes
        self.dynamic_batching = dynamic_batching
        self.num_model_instances = num_model_instances
        self.max_batch_size = max_batch_size
        self.use_trt_accelerator = use_trt_accelerator
    
        self.is_yolo = is_yolo or self.runtime_framework == "yolo"
        self.input_name = "images" if self.is_yolo else "input"
        self.output_name = "output0" if self.is_yolo else "output"

        self.gpus_count = torch.cuda.device_count()
        self.config_params = {}
        
        logging.info("Model name: %s", self.model_name)
        logging.info("Model path: %s", self.model_path)
        logging.info("Runtime framework: %s", self.runtime_framework)
        logging.info("Using connection protocol: %s", self.connection_protocol)
        logging.info("Input size: %s", self.input_size)
        logging.info("Number of classes: %s", self.num_classes)
        logging.info("Found %s GPUs available for inference", self.gpus_count)
        
        self.docker_pull_process = pull_docker_image(TRITON_DOCKER_IMAGE)

    def check_triton_docker_image(self):
        """Check if docker image download is complete and wait for it to finish"""
        logging.info("Checking docker image download status")
        stdout, stderr = self.docker_pull_process.communicate()
        if self.docker_pull_process.returncode == 0:
            logging.info(f"Docker image {TRITON_DOCKER_IMAGE} downloaded successfully")
        else:
            error_msg = stderr.decode()
            logging.error(f"Docker pull failed with return code {self.docker_pull_process.returncode}")
            logging.error("Error message: %s", error_msg)
            raise RuntimeError(f"Docker pull failed: {error_msg}")

    def prepare_model(self, model_version_dir: str) -> None:
        """Prepare the model file for Triton Inference Server.

        Copies model from self.model_path to model_version_dir and converts if necessary 
        to the format expected by Triton (model.onnx, model.xml, model.plan).

        Args:
            model_version_dir: Directory to store the model file 
            (e.g., '/models/<model_name>/1').
        """
        try:
            runtime_framework = self.runtime_framework
            logging.info("Preparing model with runtime framework: %s", runtime_framework)
            logging.info("Source model path: %s", self.model_path)

            if runtime_framework not in ["onnx", "pytorch", "torchscript", "yolo", "tensorrt", "openvino"]:
                logging.error("Runtime framework '%s' not supported. Supported: %s", 
                            runtime_framework, ["onnx", "pytorch", "torchscript", "yolo", "tensorrt", "openvino"])
                raise ValueError(f"Unsupported runtime framework: {runtime_framework}")
            
            os.makedirs(model_version_dir, exist_ok=True)
            
            # 1. ONNX - copy to model.onnx
            if runtime_framework == "onnx":
                model_file = os.path.join(model_version_dir, "model.onnx")
                if os.path.abspath(self.model_path) != os.path.abspath(model_file):
                    shutil.copy2(self.model_path, model_file)
                    logging.info("Copied ONNX model to: %s", model_file)
                else:
                    logging.info("Model path is already correct: %s", model_file)
                self._verify_onnx_model(model_file)

            # 2. PyTorch/TorchScript/Yolo - export to ONNX
            elif runtime_framework in ["pytorch", "torchscript", "yolo"]:
                model_file = os.path.join(model_version_dir, "model.onnx")
                img_chw = (3, self.input_size[0], self.input_size[1])
                logging.info("Converting %s model to ONNX with input shape: %s", runtime_framework, img_chw)
                self.to_onnx(self.model_path, model_file, (1, *img_chw))
                logging.info("Exported ONNX model to: %s", model_file)
                self._verify_onnx_model(model_file)
                
            # 3. TensorRT - copy to model.plan
            elif runtime_framework == "tensorrt":
                model_file = os.path.join(model_version_dir, "model.plan") 
                shutil.copy2(self.model_path, model_file)
                logging.info("Copied TensorRT model to: %s", model_file)
                self.runtime_framework = "tensorrt" 
                
            # 4. OpenVINO - extract ZIP or copy files
            elif runtime_framework == "openvino":
                if self.model_path.endswith('.zip'):
                    logging.info("Extracting OpenVINO ZIP to: %s", model_version_dir)
                    with zipfile.ZipFile(self.model_path, "r") as zip_ref:
                        zip_ref.extractall(model_version_dir)
                    model_file = os.path.join(model_version_dir, "model.xml")
                    model_bin_file = os.path.join(model_version_dir, "model.bin")
                else:
                    model_file = os.path.join(model_version_dir, "model.xml")
                    model_bin_file = os.path.join(model_version_dir, "model.bin")
                    
                    shutil.copy2(self.model_path, model_file)
                    
                    source_bin = self.model_path.replace('.xml', '.bin')
                    if os.path.exists(source_bin):
                        shutil.copy2(source_bin, model_bin_file)
                    else:
                        raise RuntimeError(f"OpenVINO model.bin not found at {source_bin}")

                if not os.path.exists(model_file):
                    logging.error("OpenVINO model.xml not found at %s", model_file)
                    raise RuntimeError(f"OpenVINO model.xml not found at {model_file}")
                if not os.path.exists(model_bin_file):
                    logging.error("OpenVINO model.bin not found at %s", model_bin_file)
                    raise RuntimeError(f"OpenVINO model.bin not found at {model_bin_file}")
                logging.info("Prepared OpenVINO model: %s", model_file)

            logging.info("Model preparation completed successfully")
        except Exception as e:
            logging.error("Model preparation failed: %s", str(e), exc_info=True)
            raise

    def to_onnx(self, checkpoint_path: str, onnx_path: str, input_shape: Tuple[int, int, int, int]) -> None:
        """Export PyTorch or YOLO checkpoint to ONNX."""
        try:
            runtime_framework = self.runtime_framework.lower()
            logging.info("Exporting %s model to ONNX on CPU", runtime_framework)

            if runtime_framework == "yolo":
                try:
                    if dependencies_check("ultralytics"):
                        from ultralytics import YOLO
                        logging.info("Using Ultralytics YOLO for ONNX export")
                        model = YOLO(checkpoint_path)
                        # NOTE: Update 4 -- opset=12 for YOLOv8 compatibility
                        export_path = model.export(format="onnx", imgsz=input_shape[2], dynamic=True, opset=12)  # Added opset=12
                        if export_path != onnx_path:
                            shutil.move(export_path, onnx_path)
                        logging.info("Exported YOLO model to ONNX: %s", onnx_path)
                        return
                    else:
                        logging.warning("Ultralytics not available; falling back to PyTorch export for YOLO")
                except Exception as e:
                    logging.warning("Ultralytics YOLO export failed: %s; trying PyTorch export", str(e))

            import torch
            model = torch.load(checkpoint_path, map_location="cpu")
            # TODO: Add support for model_state_dict 
            model.eval()
            dummy_input = torch.randn(*input_shape)
            torch.onnx.export(
                model,
                dummy_input,
                onnx_path,
                opset_version=17,
                input_names=["input__0"],
                output_names=["output__0"],
                dynamic_axes={"input__0": {0: "batch"}, "output__0": {0: "batch"}},
            )
            logging.info("Exported PyTorch model to ONNX: %s", onnx_path)
        except Exception as e:
            logging.error("Failed to export to ONNX: %s", str(e), exc_info=True)
            raise

    def _verify_onnx_model(self, onnx_path: str):
        """Verify that the ONNX model is valid"""
        try:
            if dependencies_check("onnx"):
                import onnx
                model = onnx.load(onnx_path)
                onnx.checker.check_model(model)
                logging.info(f"ONNX model verification successful: {onnx_path}")
            else:
                logging.warning("ONNX library not available for model verification")
        except Exception as e:
            logging.error(f"ONNX model verification failed: %s", str(e))
            raise ValueError(f"Invalid ONNX model at {onnx_path}: {str(e)}")
        
    def create_model_repository(self):
        """Create the model repository directory structure"""
        try:
            model_version = "1"
            model_dir = os.path.join(BASE_PATH, self.model_name)
            version_dir = os.path.join(model_dir, str(model_version))
            logging.info("Creating model repository structure:")
            logging.info("Base path: %s", BASE_PATH)
            logging.info("Model directory: %s", model_dir)
            logging.info("Version directory: %s", version_dir)
            os.makedirs(version_dir, exist_ok=True)
            logging.info("Model repository directories created successfully")
            return model_dir, version_dir
        except Exception as e:
            logging.error(
                "Failed to create model repository: %s",
                str(e),
                exc_info=True,
            )
            raise

    def write_config_file(
        self,
        model_dir: str,
        max_batch_size: int = 8,
        num_model_instances: int = 1,
        image_size: List[int] = [224, 224],
        num_classes: int = 10,
        input_data_type: str = "TYPE_FP32",
        output_data_type: str = "TYPE_FP32",
        dynamic_batching: bool = False,
        preferred_batch_size: list = [2, 4, 8],
        max_queue_delay_microseconds: int = 100,
        input_pinned_memory: bool = True,
        output_pinned_memory: bool = True,
        **kwargs,
    ):
        """Write the model configuration file for Triton Inference Server."""
        try:
            runtime_framework = self.runtime_framework.lower()
            logging.info("Starting to write Triton config file for framework: %s", runtime_framework)
            
            if runtime_framework == "tensorrt":
                platform = "tensorrt_plan"
                model_filename = "model.plan" 
            elif runtime_framework in ["pytorch", "torchscript", "yolo", "onnx"]:
                platform = "onnxruntime_onnx"
                model_filename = "model.onnx"
            else:
                platform = "openvino"
                model_filename = "model.xml"
            logging.info("Using %s backend with model file: %s", platform, model_filename)

            config_path = os.path.join(model_dir, "config.pbtxt")
            logging.info("Writing config to: %s", config_path)

            if platform == "onnxruntime_onnx":
                model_file = os.path.join(model_dir, "1", "model.onnx")
                if os.path.exists(model_file) and dependencies_check("onnx"):
                    import onnx
                    model = onnx.load(model_file)
                    graph = model.graph
                    
                    inputs = []
                    for inp in graph.input:
                        shape = [d.dim_value if d.HasField("dim_value") else -1 for d in inp.type.tensor_type.shape.dim]
                        inputs.append((inp.name, input_data_type, shape))
                    
                    outputs = []
                    for out in graph.output:
                        shape = [d.dim_value if d.HasField("dim_value") else -1 for d in out.type.tensor_type.shape.dim]
                        if self.is_yolo and out.name != "output__0":
                            continue  # Skip intermediate YOLO outputs
                        output_shape = shape[1:] if max_batch_size > 0 and len(shape) > 1 else shape
                        outputs.append((out.name, output_data_type, output_shape))
                    
                    logging.info("ONNX inputs: %s", inputs)
                    logging.info("ONNX outputs: %s", outputs)
                else:
                    # Fallback when ONNX / model 404
                    inputs = [(self.input_name, input_data_type, [3, image_size[0], image_size[1]])]
                    outputs = [(self.output_name, output_data_type, [-1, -1] if self.is_yolo else [num_classes])]
            elif platform == "tensorrt_plan":
                # TensorRT: Use YOLO-specific or fallback input/output configuration
                inputs = [(self.input_name, input_data_type, [3, image_size[0], image_size[1]])]
                outputs = [(self.output_name, output_data_type, [-1, -1] if self.is_yolo else [num_classes])]
            else:
                # OpenVINO fallback
                inputs = [(self.input_name, input_data_type, [3, image_size[0], image_size[1]])]
                outputs = [(self.output_name, output_data_type, [num_classes])]

            logging.info("Final inputs for config: %s", inputs)
            logging.info("Final outputs for config: %s", outputs)

            config_content = f'name: "{self.model_name}"\n'
            config_content += f'platform: "{platform}"\n'
            config_content += f'max_batch_size: {max_batch_size}\n'
            
            # Input section
            config_content += 'input [\n'
            for name, dtype, shape in inputs:
                config_content += '  {\n'
                config_content += f'    name: "{name}"\n'
                config_content += f'    data_type: {dtype}\n'
                if platform == "tensorrt_plan":
                    dims = shape
                else:
                    dims = shape[1:] if max_batch_size > 0 and len(shape) > 1 else shape
                config_content += f'    dims: [{", ".join(str(dim) for dim in dims)}]\n'
                config_content += '  }\n'
            config_content += ']\n'
            
            # Output section
            config_content += 'output [\n'
            for name, dtype, shape in outputs:
                config_content += '  {\n'
                config_content += f'    name: "{name}"\n'
                config_content += f'    data_type: {dtype}\n'
                config_content += f'    dims: [{", ".join(str(dim) for dim in shape)}]\n'
                config_content += '  }\n'
            config_content += ']\n'

            # Instance group
            if num_model_instances > 1 or self.gpus_count > 0:
                device_type = "KIND_GPU" if self.gpus_count > 0 else "KIND_CPU"
                logging.info("Adding instance group configuration for %s %s instances", num_model_instances, device_type)
                config_content += 'instance_group [\n'
                config_content += '  {\n'
                config_content += f'    count: {num_model_instances}\n'
                config_content += f'    kind: {device_type}\n'
                config_content += '  }\n'
                config_content += ']\n'

            # Dynamic batching
            if dynamic_batching:
                logging.info("Adding dynamic batching configuration")
                valid_pref_sizes = [bs for bs in preferred_batch_size if bs <= max_batch_size]
                if valid_pref_sizes:
                    config_content += 'dynamic_batching {\n'
                    config_content += f'  preferred_batch_size: [{", ".join(str(bs) for bs in valid_pref_sizes)}]\n'
                    config_content += f'  max_queue_delay_microseconds: {max_queue_delay_microseconds}\n'
                    config_content += '}\n'

            # Optimization settings
            if input_pinned_memory or output_pinned_memory or self.gpus_count > 0:
                config_content += 'optimization {\n'
                if input_pinned_memory:
                    config_content += '  input_pinned_memory {\n'
                    config_content += '    enable: true\n'
                    config_content += '  }\n'
                if output_pinned_memory:
                    config_content += '  output_pinned_memory {\n'
                    config_content += '    enable: true\n'
                    config_content += '  }\n'
                if self.gpus_count > 0 and self.use_trt_accelerator:
                    config_content += '  execution_accelerators {\n'
                    config_content += '    gpu_execution_accelerator {\n'
                    config_content += '      name: "tensorrt"\n'
                    config_content += '      parameters {\n'
                    config_content += '        key: "precision_mode"\n'
                    config_content += '        value: "FP16"\n'
                    config_content += '      }\n'
                    config_content += '      parameters {\n'
                    config_content += '        key: "max_workspace_size_bytes"\n'
                    config_content += '        value: "1073741824"\n'
                    config_content += '      }\n'
                    config_content += '      parameters {\n'
                    config_content += '        key: "trt_engine_cache_enable"\n'
                    config_content += '        value: "1"\n'
                    config_content += '      }\n'
                    config_content += '      parameters {\n'
                    config_content += '        key: "trt_engine_cache_path"\n'
                    config_content += f'        value: "/models/{self.model_name}/1"\n'
                    config_content += '      }\n'
                    config_content += '    }\n'
                    config_content += '  }\n'
                config_content += '}\n'

            with open(config_path, "w") as f:
                f.write(config_content)
            
            logging.info("Config file written successfully")
            logging.info("Config content:\n%s", config_content)
            
        except Exception as e:
            logging.error("Failed to write config file: %s", str(e), exc_info=True)
            raise

    def get_config_params(self):
        """Get configuration parameters for Triton config file"""
        try:
            logging.info("Retrieving configuration parameters")
            
            logging.info("Using input size: %s", self.input_size)
            logging.info("Using number of classes: %s", self.num_classes)
            
            params = {
                "max_batch_size": self.max_batch_size,
                "num_instances": self.num_model_instances,
                "image_size": self.input_size,
                "num_classes": self.num_classes,
                "input_data_type": "TYPE_FP32",
                "output_data_type": "TYPE_FP32",
                "dynamic_batching": self.dynamic_batching,  
                "preferred_batch_size": [1, 2, 4, 8],
                "max_queue_delay_microseconds": 100,
                "input_pinned_memory": True,
                "output_pinned_memory": True,
            }
                
            logging.debug("Final configuration parameters: %s", params)
            return params
            
        except Exception as e:
            logging.error(
                "Failed to get configuration parameters: %s",
                str(e),
                exc_info=True,
            )
            raise


    def start_server(self, internal_port: int = 8000):
        """Start the Triton Inference Server
        
        Args:
            internal_port: Port to expose the server on
        """
        gpu_option = "--gpus=all " if self.gpus_count > 0 else ""
        logging.debug("Starting Triton server with GPU option: '%s'", gpu_option.strip())
        triton_port = 8000 if self.connection_protocol == 'rest' else 8001
        port_mapping = f"-p{internal_port}:{triton_port}"
        start_triton_server = f"docker run {gpu_option}--rm {port_mapping} -v {os.path.abspath(BASE_PATH)}:/models --label model_name={self.model_name} {TRITON_DOCKER_IMAGE} tritonserver --model-repository=/models "
        logging.info("Checking docker image download status before starting server")
        self.check_triton_docker_image()
        try:
            logging.info(
                "Starting Triton server with command: %s",
                start_triton_server,
            )
            self.process = subprocess.Popen(
                shlex.split(start_triton_server),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )

            def log_output():
                while True:
                    stdout_line = self.process.stdout.readline()
                    stderr_line = self.process.stderr.readline()
                    if stdout_line:
                        logging.info(stdout_line.strip())
                    if stderr_line:
                        logging.info(stderr_line.strip())
                    if stdout_line == "" and stderr_line == "" and self.process.poll() is not None:
                        break

            threading.Thread(target=log_output, daemon=True).start()
            logging.info(
                "Triton server started successfully on port %s",
                internal_port,
            )
            return self.process
        except Exception as e:
            logging.error(
                "Failed to start Triton server: %s",
                str(e),
                exc_info=True,
            )
            raise

    def setup(self, internal_port: int = 8000):
        """Setup the Triton server with the provided model.
        
        Args:
            internal_port: Port to expose the server on
        """
        try:
            logging.info("Beginning Triton server setup")
            logging.info("Step 1: Creating model repository")
            self.model_dir, self.version_dir = self.create_model_repository()
            logging.info("Step 2: Preparing model")
            self.prepare_model(self.version_dir)
            logging.info("Step 3: Getting configuration parameters")
            self.config_params = self.get_config_params()
            logging.info("Step 4: Writing configuration file")
            self.write_config_file(
                self.model_dir,
                **self.config_params,
            )
            logging.info("Step 5: Starting Triton server")
            self.process = self.start_server(internal_port)
            logging.info("Triton server setup completed successfully")
            return self.process
        except Exception as e:
            logging.error(
                "Triton server setup failed: %s",
                str(e),
                exc_info=True,
            )
            raise


"""Module providing inference_utils functionality for FastAPI and Triton inference."""

from PIL import Image
import httpx
import logging
from typing import Optional, Dict, Union, Any
from datetime import datetime, timezone
from io import BytesIO
import numpy as np

import numpy as np
import torch
import torchvision
from typing import Tuple, Dict, Any, Optional, Union
import logging
from PIL import Image
from io import BytesIO
import os
from datetime import datetime, timezone

class TritonInference:
    """Class for making Triton inference requests."""

    def __init__(
        self,
        server_type: str,
        model_name: str,
        internal_port: int = 80,
        internal_host: str = "localhost",
        task_type: str = "detection",
        runtime_framework: str = "onnx",
        is_yolo: bool = False,
        input_size: Union[int, List[int]] = 224,
        action_tracker: Optional[Any] = None,
    ):
        """Initialize Triton inference client.

        Args:
            server_type: Type of server (grpc/rest)
            model_name: Name of model to use
            internal_port: Port number for internal API
            internal_host: Hostname for internal API
            task_type: Type of task (e.g., detection)
            runtime_framework: Framework used for the model (e.g., onnx)
            is_yolo: Boolean indicating if the model is YOLO
            input_size: Input size for the model (int or [height, width])
            action_tracker: Optional action tracker for tracking actions
        """
        self.model_name = model_name
        self.task_type = task_type
        self.runtime_framework = runtime_framework
        self.is_yolo = is_yolo

        if self.is_yolo:
            self.input_size = 640
        else:
            self.input_size = 224
        self.action_tracker = action_tracker
        # self.input_size = self.action_tracker.get_input_size() if self.action_tracker else self.input_size

        self.data_type_mapping = {
            6: "TYPE_INT8",
            7: "TYPE_INT16",
            8: "TYPE_INT32",
            9: "TYPE_INT64",
            10: "TYPE_FP16",
            11: "TYPE_FP32",
            12: "TYPE_FP64",
        }
        self.numpy_data_type_mapping = {
            "INT8": np.int8,
            "INT16": np.int16,
            "INT32": np.int32,
            "INT64": np.int64,
            "FP16": np.float16,
            "FP32": np.float32,
            "FP64": np.float64,
        }
        self.setup_client_funcs = {
            "grpc": self._setup_grpc_client,
            "rest": self._setup_rest_client,
        }
        self.url = f"{internal_host}:{internal_port}"
        self.connection_protocol = "grpc" if "grpc" in server_type else "rest"
        self.tritonclientclass = None
        self._dependencies_check()
        self.client_info = self.setup_client_funcs[self.connection_protocol]()
        logging.info(
            "Initialized TritonClientUtils with %s protocol",
            self.connection_protocol,
        )

    def _dependencies_check(self):
        """Check and import required Triton dependencies."""
        try:
            if self.connection_protocol == "rest":
                import tritonclient.http as tritonclientclass
            else:
                import tritonclient.grpc as tritonclientclass
            self.tritonclientclass = tritonclientclass
        except ImportError as err:
            package_name = "tritonclient[http]" if self.connection_protocol == "rest" else "tritonclient[grpc]"
            logging.error(
                "Failed to import tritonclient (%s): %s. Please install with: pip install %s",
                package_name, err, package_name
            )
            raise ImportError(f"Required package {package_name} not installed: {err}")
        except Exception as err:
            logging.error(
                "Failed to import tritonclient: %s",
                err,
            )
            raise

    def _setup_rest_client(self):
        """Setup REST client and model configuration.

        Returns:
            Dictionary containing client configuration
        """
        client = self.tritonclientclass.InferenceServerClient(url=self.url)
        model_config = client.get_model_config(model_name=self.model_name, model_version="1")
        input_config = model_config["input"][0]
        input_shape = [1] + input_config["dims"]  # Prepend batch dimension
        input_obj = self.tritonclientclass.InferInput(
            input_config["name"],
            input_shape,
            input_config["data_type"].split("_")[-1],
        )
        output = self.tritonclientclass.InferRequestedOutput(model_config["output"][0]["name"])
        return {
            "client": client,
            "input": input_obj,
            "output": output,
        }

    def _setup_grpc_client(self):
        """Setup gRPC client and model configuration.

        Returns:
            Dictionary containing client configuration
        """
        client = self.tritonclientclass.InferenceServerClient(url=self.url)
        model_config = client.get_model_config(model_name=self.model_name, model_version="1")
        input_config = model_config.config.input[0]
        input_shape = [1] + list(input_config.dims)  # Prepend batch dimension
        input_obj = self.tritonclientclass.InferInput(
            input_config.name,
            input_shape,
            self.data_type_mapping[input_config.data_type].split("_")[-1],
        )
        output = self.tritonclientclass.InferRequestedOutput(model_config.config.output[0].name)
        return {
            "client": client,
            "input": input_obj,
            "output": output,
        }

    def inference(self, input_data: Union[bytes, np.ndarray]) -> np.ndarray:
        """Make a synchronous inference request.

        Args:
            input_data: Input data as bytes or stacked numpy array

        Returns:
            Model prediction as numpy array

        Raises:
            Exception: If inference fails
        """
        try:
            # If already preprocessed ndarray, make it C-contiguous FP32.
            if isinstance(input_data, np.ndarray):
                input_array = np.ascontiguousarray(input_data, dtype=np.float32)
                if input_array.ndim == 5 and input_array.shape[1] == 1:
                    # [B, 1, C, H, W] -> [B, C, H, W]
                    input_array = np.ascontiguousarray(
                        input_array.reshape(input_array.shape[0],
                                           input_array.shape[2],
                                           input_array.shape[3],
                                           input_array.shape[4]),
                        dtype=np.float32
                    )
            else:
                # -> [1, C, H, W], FP32, contiguous
                input_array = self._preprocess_input(input_data)

            # Update InferInput shape to match batch (N,C,H,W)
            self.client_info["input"].set_shape(list(input_array.shape))
            self.client_info["input"].set_data_from_numpy(input_array)

            if self.connection_protocol == "rest":
                resp = self.client_info["client"].infer(
                    model_name=self.model_name,
                    model_version="1",
                    inputs=[self.client_info["input"]],
                    outputs=[self.client_info["output"]],
                )
            else:
                resp = self.client_info["client"].infer(
                    model_name=self.model_name,
                    model_version="1",
                    inputs=[self.client_info["input"]],
                    outputs=[self.client_info["output"]],
                )
                
            return resp.as_numpy(self.client_info["output"].name())

        except Exception as err:
            logging.error("Triton inference failed: %s", err, exc_info=True)
            raise Exception(f"Triton inference failed: {err}") from err

    async def async_inference(self, input_data: Union[bytes, np.ndarray]) -> np.ndarray:
        """Make an asynchronous inference request (REST + gRPC)."""
        try:
            logging.debug("Making async inference request")

            if isinstance(input_data, np.ndarray):
                input_array = input_data
            else:
                input_array = self._preprocess_input(input_data)

            # Ensure C-contiguous
            if not input_array.flags.c_contiguous:
                input_array = np.ascontiguousarray(input_array)

            self.client_info["input"].set_shape(list(input_array.shape))
            self.client_info["input"].set_data_from_numpy(input_array)

            if self.connection_protocol == "rest":
                # REST: async_infer -> InferAsyncRequest, then block with get_result()
                resp = self.client_info["client"].async_infer(
                    model_name=self.model_name,
                    model_version="1",
                    inputs=[self.client_info["input"]],
                    outputs=[self.client_info["output"]],
                )
                result = resp.get_result()

            else:
                # gRPC: async_infer uses callback; wrap it into an awaitable Future
                loop = asyncio.get_running_loop()
                fut: asyncio.Future = loop.create_future()

                def _callback(result, error):
                    if error is not None:
                        loop.call_soon_threadsafe(fut.set_exception, error)
                    else:
                        loop.call_soon_threadsafe(fut.set_result, result)

                self.client_info["client"].async_infer(
                    model_name=self.model_name,
                    model_version="1",
                    inputs=[self.client_info["input"]],
                    outputs=[self.client_info["output"]],
                    callback=_callback,
                )
                result = await fut

            logging.debug(f"Async inference response type: {type(result)}")
            logging.info("Successfully got async inference result")

            output_array = result.as_numpy(self.client_info["output"].name())
            logging.info(f"Output shape: {output_array.shape}")
            return output_array  

        except Exception as err:
            logging.error(f"Async Triton inference failed: {err}")
            raise Exception(f"Async Triton inference failed: {err}") from err
   
    def _preprocess_input(self, input_data) -> np.ndarray:
        """Preprocess input data for YOLOv8 inference.

        Args:
            input_data: Raw input bytes or string (file path)

        Returns:
            Preprocessed numpy array ready for inference
        """
        resize_shape = [int(self.input_size), int(self.input_size)]
        input_shape = [1, 3, resize_shape[0], resize_shape[1]]

        if isinstance(input_data, str) and os.path.exists(input_data):
            with open(input_data, "rb") as f:
                input_data = f.read()

        if isinstance(input_data, bytes):
            try:
                # Image file (JPEG/PNG)
                image = Image.open(BytesIO(input_data)).convert("RGB")
            except Exception as e:
                logging.warning(f"Failed to open input data as image: {e}")
                # Handle raw RGB bytes (video frames, 640x640x3) -- YoLov8 default
                arr = np.frombuffer(input_data, dtype=np.uint8).reshape(640, 640, 3)
                image = Image.fromarray(arr, mode="RGB")
        elif isinstance(input_data, np.ndarray):
            # np.ndarray (direct frame input)
            image = Image.fromarray(input_data, mode="RGB")
        else:
            raise ValueError(f"Unsupported input_data type: {type(input_data)}")

        if self.is_yolo:
            # TODO: Verify and Generalize for Obj Det models
            logging.debug("Preprocessing input for YOLO model")
            # YOLO preprocessing: resize with letterbox padding, normalize to [0, 1]
            image, ratio, (dw, dh) = self._letterbox_resize(image, resize_shape)
            arr = np.array(image).astype(np.float32) / 255.0
            arr = arr.transpose(2, 0, 1)  # Convert to CHW
            arr = np.expand_dims(arr, axis=0)  # Add batch dimension
            self.client_info["padding_info"] = {"ratio": ratio, "dw": dw, "dh": dh}
        else:
            # Classifier preprocessing: resize directly, ImageNet normalization
            image = image.resize(resize_shape)
            arr = np.array(image).astype(np.float32) / 255.0
            arr = arr.transpose(2, 0, 1)  # Convert to CHW
            arr = np.expand_dims(arr, axis=0)  # Add batch dimension
            mean = np.array([0.485, 0.456, 0.406], dtype=np.float32).reshape(1, 3, 1, 1)
            std = np.array([0.229, 0.224, 0.225], dtype=np.float32).reshape(1, 3, 1, 1)
            arr = (arr - mean) / std

        arr = arr.astype(self.numpy_data_type_mapping[self.client_info["input"].datatype()])
        logging.debug("Preprocessed input shape: %s", arr.shape)
        return arr

    def _letterbox_resize(self, image, target_size):
        """Resize image with letterbox padding to maintain aspect ratio."""
        target_h, target_w = target_size
        img_w, img_h = image.size
        ratio = min(target_w / img_w, target_h / img_h)
        new_w, new_h = int(img_w * ratio), int(img_h * ratio)
        image = image.resize((new_w, new_h), Image.LANCZOS)
        padded_image = Image.new("RGB", (target_w, target_h), (114, 114, 114))
        dw, dh = (target_w - new_w) // 2, (target_h - new_h) // 2
        padded_image.paste(image, (dw, dh))
        logging.debug("Letterbox resize: original size %s, new size %s, padding (dw, dh) (%d, %d)", (img_w, img_h), (new_w, new_h), dw, dh)
        logging.debug("Letterbox resize completed for target size: %s", target_size)
        return padded_image, ratio, (dw, dh)

    def _postprocess_yolo(
        self,
        outputs: np.ndarray,
        conf_thres: float = 0.25,
        iou_thres: float = 0.45,
        max_det: int = 300
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Postprocess YOLOv8 outputs (Torch or ONNX/Triton) with pipeline compatibility.

        Args:
            outputs: Raw model output as NumPy array, expected shape [batch, num_boxes, num_classes + 4] for YOLO
            conf_thres: Confidence threshold for filtering detections
            iou_thres: IoU threshold for Non-Maximum Suppression
            max_det: Maximum number of detections to keep

        Returns:
            Tuple of (boxes, scores, class_ids) as NumPy arrays:
            - boxes: [N, 4] array of bounding boxes in xyxy format
            - scores: [N] array of confidence scores
            - class_ids: [N] array of class IDs
        """
        if not self.is_yolo:
            return outputs, np.array([]), np.array([])

        try:
            if isinstance(outputs, np.ndarray):
                outputs = torch.from_numpy(outputs)
            elif not isinstance(outputs, torch.Tensor):
                outputs = torch.tensor(outputs, dtype=torch.float32)

            if outputs.ndim == 2:
                outputs = outputs.unsqueeze(0)
                
            if outputs.shape[1] < outputs.shape[2]:
                # Format [batch, num_classes + 4, num_boxes] -> [batch, num_boxes, num_classes + 4]
                outputs = outputs.transpose(1, 2)  

            boxes = outputs[..., :4]  # xywh, [batch, num_boxes, 4]
            scores_all = outputs[..., 4:]  # class scores, [batch, num_boxes, num_classes]

            all_boxes, all_scores, all_class_ids = [], [], []
            for batch_idx in range(outputs.shape[0]):
                batch_boxes = boxes[batch_idx]  # [num_boxes, 4]
                batch_scores_all = scores_all[batch_idx]  # [num_boxes, num_classes]

                # Get best class per box
                scores, class_ids = batch_scores_all.max(dim=-1)  

                # Confidence filter
                mask = scores > conf_thres
                batch_boxes = batch_boxes[mask]
                batch_scores = scores[mask]
                batch_class_ids = class_ids[mask]

                # Convert xywh -> xyxy
                if batch_boxes.shape[0] > 0:
                    xyxy = torch.zeros_like(batch_boxes)
                    xyxy[:, 0] = batch_boxes[:, 0] - batch_boxes[:, 2] / 2  # x1
                    xyxy[:, 1] = batch_boxes[:, 1] - batch_boxes[:, 3] / 2  # y1
                    xyxy[:, 2] = batch_boxes[:, 0] + batch_boxes[:, 2] / 2  # x2
                    xyxy[:, 3] = batch_boxes[:, 1] + batch_boxes[:, 3] / 2  # y2
                    batch_boxes = xyxy

                    # Adjust for letterbox padding if provided
                    padding_info = self.client_info.get("padding_info", {})
                    if padding_info:
                        ratio = padding_info.get("ratio", 1.0)
                        dw = padding_info.get("dw", 0)
                        dh = padding_info.get("dh", 0)
                        batch_boxes[:, 0] = (batch_boxes[:, 0] - dw) / ratio  # x1
                        batch_boxes[:, 1] = (batch_boxes[:, 1] - dh) / ratio  # y1
                        batch_boxes[:, 2] = (batch_boxes[:, 2] - dw) / ratio  # x2
                        batch_boxes[:, 3] = (batch_boxes[:, 3] - dh) / ratio  # y2

                    # NMS
                    if batch_boxes.shape[0] > 0:
                        keep = torchvision.ops.nms(batch_boxes, batch_scores, iou_thres)
                        batch_boxes = batch_boxes[keep]
                        batch_scores = batch_scores[keep]
                        batch_class_ids = batch_class_ids[keep]

                    if batch_boxes.shape[0] > max_det:
                        topk = batch_scores.topk(max_det).indices
                        batch_boxes = batch_boxes[topk]
                        batch_scores = batch_scores[topk]
                        batch_class_ids = batch_class_ids[topk]

                all_boxes.append(batch_boxes.cpu().numpy())
                all_scores.append(batch_scores.cpu().numpy())
                all_class_ids.append(batch_class_ids.cpu().numpy())

            boxes = np.concatenate(all_boxes, axis=0) if all_boxes else np.empty((0, 4))
            scores = np.concatenate(all_scores, axis=0) if all_scores else np.empty(0)
            class_ids = np.concatenate(all_class_ids, axis=0) if all_class_ids else np.empty(0)

            return boxes, scores, class_ids

        except Exception as e:
            logging.error("YOLO post-processing failed: %s", str(e), exc_info=True)
            return np.empty((0, 4)), np.empty(0), np.empty(0)

    def format_response(self, response: np.ndarray) -> Dict[str, Any]:
        """Format model response for consistent logging.

        Args:
            response: Raw model output

        Returns:
            Formatted response dictionary
        """
        if self.is_yolo:
            boxes, scores, class_ids = self._postprocess_yolo(
                response,
                conf_thres=0.25,
                iou_thres=0.45,
                max_det=300
            )
            predictions = {
                "boxes": boxes.tolist(),
                "scores": scores.tolist(),
                "class_ids": class_ids.tolist()
            }
        else:
            predictions = response.tolist() if isinstance(response, np.ndarray) else response

        return {
            "predictions": predictions,
            "model_id": self.model_name,
            "timestamp": datetime.now(timezone.utc).isoformat(),
        }


import logging
import json
import sys
from typing import Dict, Any

# Import your pipeline modules
from .data_generation import (
    create_synthetic_generation_pipeline,
    parse_synthetic_pipeline_config,
    SyntheticPipelineConfig,
    PromptStep,
)

def create_sample_pipeline_config() -> Dict[str, Any]:
    """Create a sample pipeline configuration for testing"""
    config_data = {
        "TotalNewImages": 15,
        "generationId": "gen_123",
        "dataset_id": "6800f1ab79ee9ad0a1bc8749",
        "prompts": [
            ["A futuristic cityscape at sunset"],
            ["A serene beach with palm trees"],
            ["A medieval castle under moonlight"],
            ["A dense forest with glowing fireflies"],
            ["A vibrant marketplace in an ancient town"],
            ["A snowy mountain peak at dawn"],
            ["A desert oasis with a clear blue spring"],
            ["A bustling urban street with neon lights"],
            ["A tranquil lake surrounded by mountains"],
            ["A gothic cathedral in the rain"],
            ["A tropical jungle with colorful birds"],
            ["A quiet countryside with rolling hills"],
            ["A futuristic robot in a high-tech lab"],
            ["A vintage car on a coastal road"],
            ["A starry night sky over a desert"]
        ],
        "autoGeneratePrompt": False,
        "categories": ["nature", "urban", "historical", "futuristic"]
    }
    return config_data

def create_kafka_config() -> Dict[str, Any]:
    """Create Kafka configuration with hardcoded values"""
    return {
        # Kafka broker configuration
        "bootstrap_servers": ["35.238.188.103:9092"],
        "Kafka_ip": "",
        "Kafka_port": "",
        # Topics for dataset pagination
        "dataset_request_topic": "synthetic_dataset_item_request",
        "dataset_response_topic": "synthetic_dataset_item_response",
        # Topics for upload URL management
        "upload_url_request_topic": "synthetic_PreSignedURL_request",
        "upload_url_response_topic": "synthetic_PreSignedURL_response",
        # Topic for publishing new generated dataset items
        "new_items_topic": "update_synthetic_dataset_items",
        # Consumer groups
        "consumer_group": "synthetic_pipeline_consumer",
        "upload_url_consumer_group": "upload_url_consumer",
        # Pagination settings - conservative to prevent over-consumption
        "page_size": 5  # Small page size to prevent over-queuing
    }

def create_completion_api_config() -> str:
    """Create completion API URL configuration"""
    import os
    return os.getenv('COMPLETION_API_URL', None)  # Returns None if not set, which will skip API calls

def initialize_and_run_pipeline():
    """Initialize and run the synthetic generation pipeline"""
    try:
        logging.info("=== Starting Synthetic Generation Pipeline ===")
        
        # Create sample configuration
        config_data = create_sample_pipeline_config()
        logging.info(f"Created pipeline configuration: {json.dumps(config_data, indent=2)}")
        
        # Parse pipeline configuration
        pipeline_config = parse_synthetic_pipeline_config(config_data)
        logging.info(f"Parsed pipeline config - Total images: {pipeline_config.total_new_images}")
        logging.info(f"Generation ID: {pipeline_config.generation_id}")
        logging.info(f"Dataset ID: {pipeline_config.dataset_id}")
        
        # Log prompt details
        logging.info(f"Number of prompts: {len(pipeline_config.prompt_pool)}")
        for i, prompt in enumerate(pipeline_config.prompt_pool):
            logging.info(f"  Prompt {i}: {prompt.prompt}")
        
        # Create Kafka configuration
        kafka_config = create_kafka_config()
        logging.info(f"Kafka configuration: {json.dumps(kafka_config, indent=2)}")
        
        # Completion API URL (configurable)
        completion_api_url = create_completion_api_config()
        if completion_api_url:
            logging.info(f"Using completion API URL: {completion_api_url}")
        else:
            logging.info("No completion API URL configured - will skip API calls")
        
        # Create the pipeline
        logging.info("Creating synthetic generation pipeline...")
        pipeline = create_synthetic_generation_pipeline(
            pipeline_config=pipeline_config,
            kafka_config=kafka_config
        )
        
        if pipeline is None:
            logging.error("Failed to create pipeline")
            return False
        
        logging.info("Pipeline created successfully!")
        logging.info("Pipeline stages:")
        for i, stage in enumerate(pipeline.stages):
            logging.info(f"  Stage {i}: {stage}")
        
        # Start the pipeline
        logging.info("Starting pipeline execution...")
        pipeline.start()
        
        logging.info("=== Pipeline execution completed successfully ===")
        return True
        
    except KeyboardInterrupt:
        logging.info("Pipeline execution interrupted by user")
        return False
    except Exception as e:
        logging.error(f"Error during pipeline execution: {e}")
        import traceback
        traceback.print_exc()
        return False

def run_with_custom_config(config_file_path: str = None):
    """Run pipeline with custom configuration from file"""
    try:
        if config_file_path:
            logging.info(f"Loading configuration from: {config_file_path}")
            with open(config_file_path, 'r') as f:
                config_data = json.load(f)
        else:
            logging.info("Using default sample configuration")
            config_data = create_sample_pipeline_config()
        
        # Parse and run
        pipeline_config = parse_synthetic_pipeline_config(config_data)
        kafka_config = create_kafka_config()
        completion_api_url = create_completion_api_config()
        
        if completion_api_url:
            logging.info(f"Using completion API URL: {completion_api_url}")
        else:
            logging.info("No completion API URL configured - will skip API calls")
        
        # Create and start pipeline
        pipeline = create_synthetic_generation_pipeline(
            pipeline_config=pipeline_config,
            kafka_config=kafka_config
        )
        
        if pipeline:
            logging.info("Starting pipeline execution...")
            pipeline.start()
            logging.info("Pipeline execution completed successfully")
            return True
        else:
            logging.error("Failed to create pipeline")
            return False
            
    except Exception as e:
        logging.error(f"Error running pipeline with custom config: {e}")
        import traceback
        traceback.print_exc()
        return False

def transform_synthetic_data(input_dict):
    """
    Transform input synthetic dictionary into the expected configuration format
    """

    result = {
        "TotalNewImages": input_dict.get("TotalNewImages", 0),
        "generationId": input_dict.get("_idDataGeneration", "unknown"),
        "dataset_id": input_dict.get("dataset_id", ""),
        "prompts": input_dict.get("prompts", []),
        "autoGeneratePrompt": input_dict.get("autoGeneratePrompt", False),
        "categories": input_dict.get("categories", [])
    }
    return result

class DataGeneration:
    """Class to handle dataset preparation for synthetic generation."""

    def __init__(self, session: Any, action_record_id: str):
        """Initialize DataGeneration.

        Args:
            session: Session object with RPC client
            action_record_id: ID of action record
        """
        self.session = session
        self.rpc = session.rpc
        self.action_record_id = action_record_id

        # Fetch action details
        url = f"/v1/project/action/{self.action_record_id}/details"
        response = self.rpc.get(url)
        self.action_doc = response["data"]

        # Extract fields
        self.action_type = self.action_doc.get("action")
        self.job_params = self.action_doc.get("jobParams", {})
        self.action_details = self.action_doc.get("actionDetails", {})

        logging.info(f"Job parameters: {self.job_params}")

        # Extract nested fields
        self.dataset_id = self.action_details.get("_idDataset")
        self.source_dataset_version = self.job_params.get("source_version", "")
        self.target_dataset_version = self.job_params.get("target_version", "")
        self.generation_params = self.job_params.get("generationConfig", {})

        # Transform using updated function
        self.formatted_generation = transform_synthetic_data(self.generation_params)

        self.kafka_config = self.job_params.get("kafka_config", {})

    def update_status(
        self,
        stepCode: str,
        status: str,
        status_description: str,
    ) -> None:
        """Update status of data preparation.

        Args:
            stepCode: Code indicating current step
            status: Status of step
            status_description: Description of status
        """
        try:
            logging.info(status_description)
            url = "/v1/actions"
            payload = {
                "_id": self.action_record_id,
                "action": self.action_type,
                "serviceName": self.action_doc["serviceName"],
                "stepCode": stepCode,
                "status": status,
                "statusDescription": status_description,
            }
            self.rpc.put(path=url, payload=payload)
        except Exception as e:
            logging.error(
                "Exception in update_status: %s",
                str(e),
            )
            raise

    def start_processing(self) -> None:
        """Start dataset synthetic generation processing."""
        self.update_status(
            "DCKR_PROC",
            "OK",
            "Dataset started synthetic generation",
        )

        try:
            logging.info(f"Processing synthetic generation config: {json.dumps(self.formatted_generation, indent=2)}")
            
            self.pipeline_config = parse_synthetic_pipeline_config(
                self.formatted_generation, 
                self.source_dataset_version, 
                self.target_dataset_version
            )
            logging.debug(f"Parsed pipeline config: {self.pipeline_config}")
            for attr, value in vars(self.pipeline_config).items():
                logging.info(f"Pipeline Config Attribute - {attr}: {value}")
            
            self.pipeline = create_synthetic_generation_pipeline(
                pipeline_config=self.pipeline_config,
                kafka_config=self.kafka_config,
                rpc=self.rpc,
            )
            self.pipeline.start()
            self.pipeline.wait_to_finish_processing_and_stop()
            new_dataset_items = self.pipeline.get_all_items_from_last_stage()

            try:
                completion_url = f"/v1/dataset_item/complete_bulk_data_generation/{self.pipeline_config.generation_id}"
                logging.debug(f"Calling completion API at {completion_url}")
                response = self.rpc.post(path=completion_url, payload={})

                if response and response.get("success"):
                    logging.info("Completion API called successfully")
                else:
                    logging.warning(f"Completion API call returned status {response}")
            except Exception as e:
                logging.error(f"Error calling completion API: {e}")

            self.update_status(
                "SUCCESS",
                "SUCCESS",
                "Dataset Synthetic Generation completed",
            )
        except Exception as e:
            logging.error(f"Error processing synthetic generation: {e}")
            self.update_status(
                "FAILED",
                "FAILED",
                f"Dataset Synthetic Generation failed: {str(e)}",
            )
            raise


import logging
import threading
import time
import base64
import io
from typing import Optional
from dataclasses import dataclass

import uvicorn
import httpx
import cv2
import numpy as np
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
from PIL import Image
import torch
from diffusers import StableDiffusionImg2ImgPipeline

# Configure logging
logging.basicConfig(level=logging.INFO)

@dataclass
class ImageGenerationConfig:
    """Represents image generation parameters."""
    prompt: str
    strength: float = 0.75
    guidance_scale: float = 7.5
    output_size: tuple = (768, 512)

class ImageGenerationRequest(BaseModel):
    """Request model for image generation endpoint."""
    image_url: Optional[str] = None
    image_base64: Optional[str] = None
    generation_config: ImageGenerationConfig

class ImageGenerationResponse(BaseModel):
    """Response model for image generation endpoint."""
    success: bool
    message: str
    generated_image_path: Optional[str] = None

class ImageGenerationServer:
    """Class to handle image generation server for local testing."""

    def __init__(self, port: int = 8000):
        self.port = port
        print(f"Initializing local image generation server on port: {self.port}")
        self._shutdown_event = threading.Event()
        self._server_thread = None
        self.app = FastAPI(title="Local Image Generation Server", version="1.0.0")

        # Track last request time for idle timeout
        self.last_request_time = time.time()
        self.idle_timeout_seconds = 600  # 10 minutes

        # Initialize Stable Diffusion pipeline
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        model_id = "runwayml/stable-diffusion-v1-5"
        try:
            self.pipe = StableDiffusionImg2ImgPipeline.from_pretrained(
                model_id, torch_dtype=torch.float16 if self.device == "cuda" else torch.float32
            )
            self.pipe = self.pipe.to(self.device)
            logging.info("Stable Diffusion pipeline initialized successfully")
        except Exception as e:
            logging.error(f"Failed to initialize Stable Diffusion pipeline: {e}")
            raise

        self._setup_routes()
        self._start_idle_monitor()

        logging.info("Successfully initialized local image generation server on port: %d", self.port)

    def _start_idle_monitor(self):
        """Background thread that shuts down server if idle for too long."""
        def monitor():
            while not self._shutdown_event.is_set():
                time.sleep(60)  # Check every 1 minute
                idle_time = time.time() - self.last_request_time
                if idle_time > self.idle_timeout_seconds:
                    logging.info("No requests received for 10 minutes. Stopping server...")
                    self.stop_server()
                    break
        t = threading.Thread(target=monitor, daemon=True, name="IdleMonitor")
        t.start()

    def _generate_image(self, image_np: np.ndarray, config: ImageGenerationConfig) -> np.ndarray:
        """Generate a new image using Stable Diffusion."""
        try:
            # Convert numpy array to PIL Image
            image_pil = Image.fromarray(image_np).convert("RGB")
            # Resize to specified output size
            image_pil = image_pil.resize(config.output_size)

            # Generate image using Stable Diffusion
            generated_images = self.pipe(
                prompt=config.prompt,
                image=image_pil,
                strength=config.strength,
                guidance_scale=config.guidance_scale
            ).images
            generated_image = generated_images[0]

            # Convert back to numpy array
            generated_image_np = np.array(generated_image)
            return generated_image_np
        except Exception as e:
            logging.error(f"Failed to generate image: {e}", exc_info=True)
            raise HTTPException(status_code=500, detail=f"Image generation failed: {str(e)}")

    def _setup_routes(self) -> None:
        """Setup FastAPI routes."""
        @self.app.post("/generate", response_model=ImageGenerationResponse)
        async def generate_image(request: ImageGenerationRequest):
            """Generate a new image based on input image and prompt."""
            self.last_request_time = time.time()  # Update last activity
            try:
                logging.info("Received image generation request: %s", request)
                if not request.image_url and not request.image_base64:
                    raise HTTPException(status_code=400, detail="Either image_url or image_base64 must be provided")
                if not request.generation_config.prompt:
                    raise HTTPException(status_code=400, detail="A prompt must be provided in generation_config")

                # Load input image
                image_pil = await self._load_image(request.image_url, request.image_base64)
                image_np = np.array(image_pil)

                # Generate new image
                generated_image_np = self._generate_image(
                    image_np=image_np,
                    config=request.generation_config
                )

                # Convert to BGR for OpenCV encoding
                image_bgr = cv2.cvtColor(generated_image_np, cv2.COLOR_RGB2BGR)
                is_success, img_encoded = cv2.imencode(".jpg", image_bgr)
                if not is_success:
                    raise HTTPException(status_code=500, detail="Failed to encode generated image")

                # Save generated image locally
                output_path = f"generated_image_{int(time.time())}.jpg"
                cv2.imwrite(output_path, image_bgr)
                logging.info(f"Generated image saved to {output_path}")

                return ImageGenerationResponse(
                    success=True,
                    message="Image generation completed and saved locally",
                    generated_image_path=output_path
                )
            except HTTPException:
                raise
            except Exception as e:
                logging.error(f"Error in image generation: {str(e)}", exc_info=True)
                raise HTTPException(status_code=500, detail=f"Internal server error: {str(e)}")

        @self.app.get("/health")
        async def health_check():
            self.last_request_time = time.time()  # Update last activity
            return {"status": "healthy", "server": "image_generation_server"}

        @self.app.get("/")
        async def root():
            self.last_request_time = time.time()  # Update last activity
            return {"message": "Local Image Generation Server", "version": "1.0.0"}

    async def _load_image(self, image_url: Optional[str], image_base64: Optional[str]) -> Image.Image:
        """Load image from URL or base64 string and return PIL Image."""
        try:
            if image_url:
                if image_url.startswith(("http://", "https://")):
                    async with httpx.AsyncClient() as client:
                        response = await client.get(image_url)
                        response.raise_for_status()
                        image = Image.open(io.BytesIO(response.content))
                else:
                    image = Image.open(image_url)
            elif image_base64:
                image_data = base64.b64decode(image_base64)
                image = Image.open(io.BytesIO(image_data))
            else:
                raise ValueError("No image source provided.")
            if image.mode != "RGB":
                image = image.convert("RGB")
            return image
        except Exception as e:
            raise HTTPException(status_code=400, detail=f"Failed to load image: {str(e)}")

    def start_server(self) -> None:
        """Start the image generation server."""
        try:
            logging.info("Starting uvicorn server on localhost:%d", self.port)
            def run_server():
                try:
                    uvicorn.run(self.app, host="0.0.0.0", port=self.port, log_level="info")
                except Exception as exc:
                    logging.error("Failed to start server: %s", str(exc))
                    raise
            self._server_thread = threading.Thread(target=run_server, daemon=False, name="ImageGenerationServer")
            self._server_thread.start()
            time.sleep(2)
            logging.info("Image generation server started successfully on localhost:%d", self.port)
        except Exception as e:
            logging.error(f"Failed to start server: {e}")
            raise

    def stop_server(self) -> None:
        """Stop the image generation server."""
        if not self._shutdown_event.is_set():
            logging.info("Stopping image generation server...")
            self._shutdown_event.set()

    def wait_for_shutdown(self) -> None:
        """Wait for the server to be shut down."""
        if self._server_thread and self._server_thread.is_alive():
            self._server_thread.join()

if __name__ == "__main__":
    server = ImageGenerationServer(port=8000)
    server.start_server()



import logging
import threading
import urllib.request
from typing import Optional, Any, Tuple, List, Dict
import signal
import atexit
import base64
import io
import time
from huggingface_hub import login
import uvicorn
import httpx
import cv2
import numpy as np
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
from PIL import Image
import torch
from diffusers import AutoPipelineForImage2Image
from diffusers.utils import make_image_grid, load_image
from diffusers import DiffusionPipeline
from diffusers.utils import load_image
from dataclasses import dataclass
from matrice.utils import dependencies_check
from matrice.utils import handle_response

dependencies_check(["httpx", "opencv-python-headless", "torch", "diffusers"])

class ImageGenerationConfig(BaseModel):
    """Represents image generation parameters."""
    prompt: str
    # strength: Optional[float] = 0.75
    # guidance_scale: Optional[float] = 7.5
    # output_size: Optional[tuple] = (768, 512)

class ImageGenerationRequest(BaseModel):
    """Request model for image generation endpoint."""
    image_url: Optional[str] = None
    image_base64: Optional[str] = None
    upload_url: str
    generation_config: ImageGenerationConfig

class ImageGenerationResponse(BaseModel):
    """Response model for image generation endpoint."""
    success: bool
    message: str
    generated_image_url: Optional[str] = None
    generated_bboxes: Optional[List[Dict]] = None

class ImageGenerationServer:
    """Class to handle image generation server."""

    def __init__(self, session: Any, action_record_id: str, port: int, ip_address: str = None):
        self.session = session
        self.rpc = session.rpc
        self.action_record_id = action_record_id
        self.port = port
        self.ip = ip_address or self._get_external_ip()
        self._shutdown_event = threading.Event()
        self._server_thread = None
        self.app = FastAPI(title="Matrice Image Generation Server", version="1.0.0")

        # Track last request time
        self.last_request_time = time.time()
        self.idle_timeout_seconds = 6000  # 10 minutes

        # Fetch Hugging Face token and log in
        try:
            hf_token = self.get_hugging_face_token_for_data_generation()
            logging.info(f"retrieved Hugging Face token: {hf_token}")
            login(token=hf_token)
            logging.info("Successfully logged in to Hugging Face with token")
        except Exception as e:
            logging.error(f"Failed to log in to Hugging Face: {e}")
            raise RuntimeError(f"Failed to authenticate with Hugging Face: {str(e)}")

        # Initialize Stable Diffusion pipeline
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        model_id = "kandinsky-community/kandinsky-2-2-decoder"
        try:
            self.pipe = AutoPipelineForImage2Image.from_pretrained(
                model_id, torch_dtype=torch.float16, use_safetensors=True
            )
            # self.pipe = DiffusionPipeline.from_pretrained(
            #     model_id, torch_dtype=torch.float16 if self.device == "cuda" else torch.float32
            # )
            # self.pipe.enable_model_cpu_offload()
            self.pipe.enable_xformers_memory_efficient_attention()
            self.pipe = self.pipe.to(self.device)
            logging.info("Stable Diffusion pipeline initialized successfully")
        except Exception as e:
            logging.error(f"Failed to initialize Stable Diffusion pipeline: {e}")
            raise

        self._setup_routes()
        self._setup_shutdown_handlers()
        
        self._fetch_action_details()
        self._start_idle_monitor()

        logging.info("Successfully initialized image generation server on IP: %s", self.ip)

    def get_hugging_face_token_for_data_generation(self):
        """Retrieve Hugging Face token for data generation."""
        secret_name = "hugging_face"
        resp, error, message = self.get_model_secret_keys(secret_name)
        if error is not None:
            logging.error("Error getting Hugging Face token: %s", message)
            raise RuntimeError(f"Failed to retrieve Hugging Face token: {message}")
        else:
            hugging_face_token = resp["user_access_token"]
            return hugging_face_token

    def get_model_secret_keys(self, secret_name):
        """Get model secret keys.

        Args:
            secret_name: Name of the secret

        Returns:
            Tuple of (data, error, message) from API response
        """
        path = f"/v1/scaling/get_models_secret_keys?secret_name={secret_name}"
        resp = self.rpc.get(path=path)
        return handle_response(
            resp,
            "Secret keys fetched successfully",
            "Could not fetch the secret keys",
        )

    def _start_idle_monitor(self):
        """Background thread that shuts down server if idle for too long."""
        def monitor():
            while not self._shutdown_event.is_set():
                time.sleep(60)  # Check every 1 minute
                idle_time = time.time() - self.last_request_time
                if idle_time > self.idle_timeout_seconds:
                    logging.info("No requests received for 10 minutes. Stopping server due to inactivity...")
                    self.stop_server()
                    break
        t = threading.Thread(target=monitor, daemon=True, name="IdleMonitor")
        t.start()

    def _generate_image(self, image_pil: Image.Image, config: ImageGenerationConfig) -> np.ndarray:
        """Generate a new image using the Qwen model."""
        try:
            # Ensure image is in RGB mode
            if image_pil.mode != "RGB":
                image_pil = image_pil.convert("RGB")    

            # Generate image using Qwen model
            generated_images = self.pipe(
                prompt=config.prompt,
                image=image_pil,
                guidance_scale=12.5,
                strength=0.4,
                num_inference_steps=50
            ).images
            generated_image = generated_images[0]

            # Convert back to numpy array for further processing
            generated_image_np = np.array(generated_image)
            return generated_image_np
        except RuntimeError as e:
            if "out of memory" in str(e).lower():
                logging.error("CUDA out of memory during image generation")
                raise HTTPException(status_code=500, detail="GPU out of memory. Try reducing image size.")
            logging.error(f"Failed to generate image: {e}", exc_info=True)
            raise HTTPException(status_code=500, detail=f"Image generation failed: {str(e)}")
        except Exception as e:
            logging.error(f"Failed to generate image: {e}", exc_info=True)
            raise HTTPException(status_code=500, detail=f"Image generation failed: {str(e)}")

    def _setup_routes(self) -> None:
        """Setup FastAPI routes."""
        @self.app.post("/generate", response_model=ImageGenerationResponse)
        async def generate_image(request: ImageGenerationRequest):
            """Generate a new image based on input image and prompt."""
            self.last_request_time = time.time()  # Update last activity
            try:
                logging.info("Received image generation request: %s", request)
                if not request.image_url and not request.image_base64:
                    raise HTTPException(status_code=400, detail="Either image_url or image_base64 must be provided")
                if not request.upload_url:
                    raise HTTPException(status_code=400, detail="An upload_url must be provided")
                if not request.generation_config.prompt:
                    raise HTTPException(status_code=400, detail="A prompt must be provided in generation_config")

                # Load input image
                if request.image_url:
                    image_pil = load_image(request.image_url)
                else:
                    image_pil = await self._load_image(None, request.image_base64)

                # Generate new image
                generated_image_np = self._generate_image(
                    image_pil=image_pil,
                    config=request.generation_config
                )

                # Convert to BGR for OpenCV encoding
                image_bgr = cv2.cvtColor(generated_image_np, cv2.COLOR_RGB2BGR)
                is_success, img_encoded = cv2.imencode(".jpg", image_bgr, [int(cv2.IMWRITE_JPEG_QUALITY), 90])
                if not is_success:
                    raise HTTPException(status_code=500, detail="Failed to encode generated image")

                # Upload generated image
                img_bytes = img_encoded.tobytes()
                async with httpx.AsyncClient() as client:
                    upload_response = await client.put(
                        request.upload_url,
                        content=img_bytes,
                        headers={'Content-Type': 'image/jpeg'},
                        timeout=60.0,
                    )
                if upload_response.status_code not in [200, 201, 204]:
                    error_detail = f"Failed to upload image. Status: {upload_response.status_code}. Response: {upload_response.text}"
                    logging.error(error_detail)
                    raise HTTPException(status_code=500, detail=error_detail)

                permanent_url = request.upload_url.split('?')[0]
                return ImageGenerationResponse(
                    success=True,
                    message="Image generation completed and uploaded successfully",
                    generated_image_url=permanent_url
                )
            except HTTPException:
                raise
            except Exception as e:
                logging.error(f"Error in image generation: {str(e)}", exc_info=True)
                raise HTTPException(status_code=500, detail=f"Internal server error: {str(e)}")

        @self.app.get("/health")
        async def health_check():
            self.last_request_time = time.time()  # Update last activity
            return {"status": "healthy", "server": "image_generation_server"}

        @self.app.get("/")
        async def root():
            self.last_request_time = time.time()  # Update last activity
            return {"message": "Matrice Image Generation Server", "version": "1.0.0"}

    def _get_external_ip(self) -> str:
        """Get external IP address."""
        try:
            with urllib.request.urlopen("https://ident.me", timeout=60) as response:
                return response.read().decode("utf8").strip()
        except Exception as e:
            logging.warning(f"Failed to get external IP: {e}, using localhost")
            return "localhost"

    def _fetch_action_details(self) -> None:
        """Fetch action details from the API."""
        try:
            url = f"/v1/project/action/{self.action_record_id}/details"
            response = self.rpc.get(url)
            self.action_doc = response["data"]
            self.action_type = self.action_doc.get("action")
            self.job_params = self.action_doc.get("jobParams", {})
            self.action_details = self.action_doc.get("actionDetails", {})
            self.generation_server_id = self.action_details.get(
                "serverId",
                self.job_params.get("serverId", None),
            )
        except Exception as e:
            logging.error(f"Failed to fetch action details: {e}")
            self.action_doc = {}
            self.action_type = "image_generation"
            self.job_params = {}
            self.action_details = {}
            self.generation_server_id = None

    async def _load_image(self, image_url: Optional[str], image_base64: Optional[str]) -> Image.Image:
        """Load image from URL or base64 string and return PIL Image."""
        try:
            if image_url:
                if image_url.startswith(("http://", "https://")):
                    async with httpx.AsyncClient() as client:
                        response = await client.get(image_url)
                        response.raise_for_status()
                        image = Image.open(io.BytesIO(response.content))
                else:
                    image = Image.open(image_url)
            elif image_base64:
                image_data = base64.b64decode(image_base64)
                image = Image.open(io.BytesIO(image_data))
            else:
                raise ValueError("No image source provided.")

            if image.mode != "RGB":
                image = image.convert("RGB")
            return image
        except Exception as e:
            raise HTTPException(status_code=400, detail=f"Failed to load image: {str(e)}")

    def _setup_shutdown_handlers(self) -> None:
        """Setup shutdown signal handlers."""
        def signal_handler(signum, frame):
            logging.info(f"Received signal {signum}, shutting down gracefully...")
            self.stop_server()
        def atexit_handler():
            try:
                self.stop_server()
            except Exception as exc:
                logging.error("Error during atexit shutdown: %s", str(exc))
        signal.signal(signal.SIGTERM, signal_handler)
        signal.signal(signal.SIGINT, signal_handler)
        atexit.register(atexit_handler)
        logging.info("Shutdown handlers registered successfully")

    def update_status(self, stepCode: str, status: str, status_description: str) -> None:
        """Update status of image generation server."""
        try:
            logging.info(status_description)
            url = "/v1/actions"
            payload = {
                "_id": self.action_record_id,
                "action": self.action_type,
                "serviceName": self.action_doc.get("serviceName", "Unknown"),
                "stepCode": stepCode,
                "status": status,
                "statusDescription": status_description,
            }
            self.rpc.put(path=url, payload=payload)
        except Exception as e:
            logging.error("Exception in update_status: %s", str(e))

    def update_server_address(self, status, port, host) -> None:
        """Update server address in the backend."""
        try:
            path = "/v1/actions/synthetic_data_servers"
            payload = {
                "id": self.generation_server_id,
                "host": host,
                "port": int(port),
                "status": status,
                "isShared": True,
                "accountNumber": self.action_doc.get("accountNumber", ""),
            }
            resp = self.rpc.put(path=path, payload=payload)
            logging.info(f"Server address update response: {resp}")
        except Exception as e:
            logging.error(f"Failed to update server address: {e}")

    def start_server(self) -> None:
        """Start the image generation server."""
        try:
            self.update_status("DCKR_PROC", "OK", "Starting image generation server")
            self.update_server_address("running", self.port, self.ip)
            def run_server():
                try:
                    logging.info("Starting uvicorn server on %s:%d", self.ip, int(self.port))
                    uvicorn.run(self.app, host="0.0.0.0", port=int(self.port), log_level="info")
                except Exception as exc:
                    logging.error("Failed to start image generation server: %s", str(exc))
                    self.update_status("ERROR", "ERROR", f"Failed to start server: {str(exc)}")
                    raise
            self._server_thread = threading.Thread(target=run_server, daemon=False, name="ImageGenerationServer")
            self._server_thread.start()
            time.sleep(2)
            self.update_status("SUCCESS", "SUCCESS", f"Image generation server started successfully on {self.ip}:{self.port}")
            logging.info("Image generation server thread started successfully")
        except Exception as e:
            logging.error(f"Failed to start server: {e}")
            self.update_status("ERROR", "ERROR", f"Failed to start image generation server: {str(e)}")
            raise

    def stop_server(self) -> None:
        """Stop the image generation server gracefully."""
        if not self._shutdown_event.is_set():
            logging.info("Stopping image generation server...")
            self._shutdown_event.set()
            try:
                self.update_status("STOPPED", "STOPPED", "Image generation server stopped")
                self.update_server_address("stopped", 0, "")
            except Exception as e:
                logging.error(f"Failed to update stop status: {e}")

    def wait_for_shutdown(self) -> None:
        """Wait for the server to be shut down."""
        if self._server_thread and self._server_thread.is_alive():
            self._server_thread.join()